import numpy as np
import sympy as sp
from typing import Protocol, Any

Matrix = np.ndarray | sp.Matrix
symbolic = sp.core.symbol.Symbol

class MatrixElement(Protocol):
    def __init__(self, value: complex | symbolic) -> None: ...
    @property
    def value(self) -> Any: ...

    @property
    def isfinite(self) -> bool: ...

class NumericMatrixElement:
    def __init__(self, value: complex | symbolic) -> None:
        try:
            self._value = complex(value)
        except ValueError:
            self._value = np.nan

    @property
    def value(self) -> complex:
        return self._value

    @property
    def isfinite(self) -> bool:
        if np.isnan(self.value):
            return True
        return np.isfinite(self.value)

class SymbolicMatrixElement:
    def __init__(self, value: complex | symbolic) -> None:
        self._value = sp.sympify(value)

    @property
    def value(self) -> Any:
        return self._value

    @property
    def isfinite(self) -> bool:
        if self.value == sp.nan:
            return True
        if self.value.is_finite is None:
            return True
        return self.value.is_finite

class MatrixOperations(Protocol):
    @staticmethod
    def zeros(shape: tuple[int, int]) -> Any: ...

    @staticmethod
    def nan(shape: tuple[int, int]) -> Any: ...

    @staticmethod
    def column_vector(values: list[complex | symbolic]) -> Any: ...

    @staticmethod
    def vstack(matrices: tuple[Any, ...]) -> Any: ...

    @staticmethod
    def hstack(matrices: tuple[Any, ...]) -> Any: ...
    
    @staticmethod
    def diag(values: list[complex | float | symbolic]) -> Any: ...

    @staticmethod
    def diag_vec(values: Any) -> list[complex | symbolic]: ...

    @staticmethod
    def inv(matrix: Any) -> Any: ...

    @staticmethod
    def solve(A: Any, b: Any) -> tuple[complex | symbolic, ...]: ...

    @staticmethod
    def elm(value: complex | symbolic) -> MatrixElement: ...

    @staticmethod
    def shape(matrix: Any) -> tuple[int, int]: ...

    @staticmethod
    def contains_nan(matrix: Any) -> bool: ...

    @staticmethod
    def any_element(matrix: Any, axis: int) -> np.ndarray: ...

    @staticmethod
    def delete(matrix: Any, idx: list[int], axis: int) -> Any: ...

class NumPyMatrixOperations:
    @staticmethod
    def zeros(shape: tuple[int, int]) -> np.ndarray:
        return np.zeros(shape, dtype=complex)

    @staticmethod
    def nan(shape: tuple[int, int]) -> np.ndarray:
        return np.full(shape, np.nan, dtype=complex)

    @staticmethod
    def column_vector(values: list[complex | symbolic]) -> Any:
        return np.array([NumericMatrixElement(v).value for v in values]).reshape(len(values), 1)

    @staticmethod
    def vstack(matrices: tuple[np.ndarray, ...]) -> np.ndarray:
        return np.vstack(matrices)

    @staticmethod
    def hstack(matrices: tuple[np.ndarray, ...]) -> np.ndarray:
        return np.hstack(matrices)

    @staticmethod
    def diag(values: list[complex | symbolic]) -> np.ndarray:
        return np.diag([NumericMatrixElement(v).value for v in values])

    @staticmethod
    def diag_vec(values: np.ndarray) -> list[complex | symbolic]:
        return [NumericMatrixElement(v).value for v in np.diag(values)]

    @staticmethod
    def inv(matrix: np.ndarray) -> np.ndarray:
        return np.linalg.inv(matrix)

    @staticmethod
    def solve(A: np.ndarray, b: np.ndarray) -> tuple[complex, ...]:
        return tuple(np.linalg.solve(A, b).flatten())

    @staticmethod
    def elm(value: complex | symbolic) -> NumericMatrixElement:
        return NumericMatrixElement(value)

    @staticmethod
    def shape(matrix: np.ndarray) -> tuple[int, int]:
        return (matrix.shape[0], matrix.shape[1])

    @staticmethod
    def contains_nan(matrix: np.ndarray) -> bool:
        return np.any(np.isnan(matrix)) == True

    @staticmethod
    def any_element(matrix: np.ndarray, axis: int) -> np.ndarray:
        return np.any(matrix, axis=axis)

    @staticmethod
    def delete(matrix: np.ndarray, idx: list[int], axis: int) -> np.ndarray:
        return np.delete(matrix, idx, axis)

class SymPyMatrixOperations:
    @staticmethod
    def zeros(shape: tuple[int, int]) -> sp.Matrix:
        return sp.zeros(*shape)

    @staticmethod
    def nan(shape: tuple[int, int]) -> sp.Matrix:
        return sp.Matrix([[sp.nan] * shape[1]] * shape[0])

    @staticmethod
    def column_vector(values: list[complex | symbolic]) -> sp.Matrix:
        return sp.Matrix([[v] for v in values]).reshape(len(values), 1)

    @staticmethod
    def vstack(matrices: tuple[sp.Matrix, ...]) -> sp.Matrix:
        return sp.Matrix.vstack(*matrices)

    @staticmethod
    def hstack(matrices: tuple[sp.Matrix, ...]) -> sp.Matrix:
        return sp.Matrix.hstack(*[sp.Matrix(m) for m in matrices])

    @staticmethod
    def diag(values: list[complex | symbolic]) -> sp.Matrix:
        return sp.diag(*values)

    @staticmethod
    def diag_vec(values: sp.Matrix) -> list[complex | symbolic]:
        return list(values.diagonal())

    @staticmethod
    def inv(matrix: sp.Matrix) -> sp.Matrix:
        return sp.Matrix(matrix.inv())

    @staticmethod
    def solve(A: sp.Matrix, b: sp.Matrix) -> tuple[symbolic, ...]:
        return tuple(A.LUsolve(b))

    @staticmethod
    def elm(value: complex | symbolic) -> SymbolicMatrixElement:
        return SymbolicMatrixElement(value)

    @staticmethod
    def shape(matrix: sp.Matrix) -> tuple[int, int]:
        return (matrix.shape[0], matrix.shape[1])

    @staticmethod
    def contains_nan(matrix: sp.Matrix) -> bool:
        return False

    @staticmethod
    def any_element(matrix: sp.Matrix, axis: int) -> np.ndarray:
        is_zero = np.array(matrix.applyfunc(lambda x: int(x != 0)), dtype=int)
        return np.any(is_zero, axis)

    @staticmethod
    def delete(matrix: sp.Matrix, idx: list[int], axis: int) -> sp.Matrix:
        if len(idx) == 0:
            return matrix
        if axis == 0:
            return matrix.row_del(*idx)
        if axis == 1:
            return matrix.col_del(*idx)
        return matrix


from typing import Protocol
from dataclasses import dataclass
from abc import ABC, abstractmethod
import numpy as np
import sympy as sp

symbolic = sp.core.symbol.Symbol

class NortenTheveninElement(Protocol):
    @property
    def name(self) -> str:
        """Name of element"""
        ...
    @property
    def type(self) -> str:
        """Specific element type"""
        ...
    @property
    def Z(self) -> complex | symbolic:
        """Impedance value of element"""
        ...
    @property
    def Y(self) -> complex | symbolic:
        """Admittance value of element"""
        ...
    @property
    def V(self) -> complex | symbolic:
        """Voltage value of element"""
        ...
    @property
    def I(self) -> complex | symbolic:
        """Current value of element"""
        ...
    @property
    def is_voltage_source(self) -> bool:
        """Check if element behaves as a linear voltage source"""
        ...
    @property
    def is_current_source(self) -> bool:
        """Check if element behaves as a linear current source"""
        ...
    @property
    def is_ideal_voltage_source(self) -> bool:
        """Check if element behaves as an ideal voltage source"""
        ...
    @property
    def is_ideal_current_source(self) -> bool:
        """Check if element behaves as an ideal current source"""
        ...
    @property
    def is_active(self) -> bool:
        """Check if element is active"""
        ...
    @property
    def is_short_circuit(self) -> bool:
        """Check if element behaves as a short circuit"""
        ... 
    @property
    def is_open_circuit(self) -> bool:
        """Check if element behaves as an open circuit"""
        ...

@dataclass(frozen=True)
class NumericNortenTheveninElement(ABC):
    name: str
    type: str

    @property
    @abstractmethod
    def Z(self) -> complex: ...

    @property
    @abstractmethod
    def Y(self) -> complex: ...

    @property
    @abstractmethod
    def V(self) -> complex: ...

    @property
    @abstractmethod
    def I(self) -> complex: ...

    @property
    def is_voltage_source(self) -> bool:
        return np.abs(self.V) > 0

    @property
    def is_current_source(self) -> bool:
        return np.abs(self.I) > 0

    @property
    def is_ideal_voltage_source(self) -> bool:
        return np.abs(self.V) >= 0 and self.Z==0

    @property
    def is_ideal_current_source(self) -> bool:
        return np.abs(self.I) >= 0 and self.Y==0

    @property
    def is_active(self) -> bool:
        return self.is_voltage_source or self.is_current_source

    @property
    def is_short_circuit(self) -> bool:
        return self.V == 0 and self.Z == 0

    @property
    def is_open_circuit(self) -> bool:
        return self.I == 0 and self.Y == 0

@dataclass(frozen=True)
class NortenElement(NumericNortenTheveninElement):
    V : complex = 0
    Z : complex = 0

    @property
    def Y(self) -> complex:
        try:
            return 1/self.Z
        except ZeroDivisionError:
            return np.inf

    @property
    def I(self) -> complex:
        try:
            return complex(self.V)/self.Z
        except ZeroDivisionError:
            return np.nan

@dataclass(frozen=True)
class TheveninElement(NumericNortenTheveninElement):
    I : complex = 0
    Y : complex = 0

    @property
    def Z(self) -> complex:
        try:
            return 1/self.Y
        except ZeroDivisionError:
            return np.inf

    @property
    def V(self) -> complex:
        try:
            return complex(self.I)/self.Y
        except ZeroDivisionError:
            return np.nan

@dataclass(frozen=True)
class SymbolicNortenTheveninElement(ABC):
    name: str
    type: str

    @property
    @abstractmethod
    def Z(self) -> symbolic: ...

    @property
    @abstractmethod
    def Y(self) -> symbolic: ...

    @property
    @abstractmethod
    def V(self) -> symbolic: ...

    @property
    @abstractmethod
    def I(self) -> symbolic: ...

    @property
    def is_voltage_source(self) -> bool:
        return self.V != 0 and self.V != sp.nan

    @property
    def is_current_source(self) -> bool:
        return self.I != 0 and self.I != sp.nan

    @property
    def is_ideal_voltage_source(self) -> bool:
        return self.V != sp.nan and self.Z == 0

    @property
    def is_ideal_current_source(self) -> bool:
        return self.I != sp.nan and self.Y == 0

    @property
    def is_active(self) -> bool:
        return self.is_voltage_source or self.is_current_source

    @property
    def is_short_circuit(self) -> bool:
        return self.V == 0 and self.Z == 0

    @property
    def is_open_circuit(self) -> bool:
        return self.I == 0 and self.Y == 0

@dataclass(frozen=True)
class SymbolicNortenElement(SymbolicNortenTheveninElement):
    V : symbolic = sp.sympify(0)
    Z : symbolic = sp.sympify(0)

    @property
    def Y(self) -> symbolic:
        if self.Z == 0:
            return sp.sympify('oo')
        if abs(self.Z) == sp.oo:
            return sp.sympify(0)
        return 1/self.Z # type: ignore
    
    @property
    def I(self) -> symbolic:
        if self.Z == 0:
            return sp.nan
        if abs(self.V) == sp.oo and abs(self.Z) == sp.oo:
            return sp.nan
        if abs(self.Z) == sp.oo:
            return sp.sympify(0)
        return self.V/self.Z # type: ignore

@dataclass(frozen=True)
class SymbolicTheveninElement(SymbolicNortenTheveninElement):
    I : symbolic = sp.sympify(0)
    Y : symbolic = sp.sympify(0)

    @property
    def Z(self) -> symbolic:
        if self.Y == 0:
            return sp.sympify('oo')
        if abs(self.Y) == sp.oo:
            return sp.sympify(0)
        return 1/self.Y # type: ignore
    
    @property
    def V(self) -> symbolic:
        if self.Y == 0:
            return sp.nan
        if abs(self.I) == sp.oo and abs(self.Y) == sp.oo:
            return sp.nan
        if abs(self.Y) == sp.oo:
            return sp.sympify(0)
        return self.I/self.Y # type: ignore


from . import layout
from ..Circuit.solution import FrequencyDomainSolution, FrequencyDomainSeries
from typing import Callable, TypedDict
import numpy as np
import functools

def plot_discrete_frequencies_fcn(ax:layout.Axis, w:np.ndarray, X:np.ndarray, **kwargs) -> None:
    line = ax.plot([w[0], w[0]], [0, X[0]], **kwargs)
    color = line[0].get_color()
    kwargs.update({'label': '', 'color': color})
    for w0, X0 in zip(w[1:], X[1:]):
        ax.plot([w0, w0], [0, X0], **kwargs)

def plot_frequencies_by_id(
        id:str,
        **kwargs) -> layout.PlotFcn:
    def plot_frequencies(
            fig:layout.Figure,
            ax:layout.Axes,
            fd_fcn:Callable[[str], FrequencyDomainSeries]=lambda _: (np.array(0), np.array(0)),
            label_fcn:Callable[[str], str]=lambda _: '',
            **_) -> layout.FigureAxes:
        if 'label' not in kwargs.keys():
            kwargs.update({'label':label_fcn(id)})
        w, X = fd_fcn(id)
        plot_discrete_frequencies_fcn(ax[-2], w, np.abs(X), **kwargs)
        plot_discrete_frequencies_fcn(ax[-1], w, np.angle(X), **kwargs)
        return fig, ax
    return plot_frequencies

def plot_frequencies_by_fcn(
        fd_fcn:FrequencyDomainSeries,
        **kwargs) -> layout.PlotFcn:
    def plot_frequencies(fig:layout.Figure, ax:layout.Axes, **_) -> layout.FigureAxes:
        plot_discrete_frequencies_fcn(ax[-2], fd_fcn[0], np.abs(fd_fcn[1]), **kwargs)
        plot_discrete_frequencies_fcn(ax[-1], fd_fcn[0], np.angle(fd_fcn[1]), **kwargs)
        return fig, ax
    return plot_frequencies

class FrequencyDomainPlotProperties(TypedDict):
    fd_fcn: Callable[[str], FrequencyDomainSeries]
    label_fcn: Callable[[str], str]
    ylabel:str

def plot_frequencydomain_factory(
        *args,
        type:str='default',
        solution:FrequencyDomainSolution,
        wmin:float=0,
        wmax:float=-1,
        abs_min:float=0,
        abs_max:float=-1,
        phi_min:float=-np.pi,
        phi_max:float=+np.pi,
        logw:bool=False,
        logy:bool=False,
        xlabel:str='ω→',
        ylabel:str='') -> layout.PlotFcn:

    fd_plot_config: dict[str, FrequencyDomainPlotProperties] = {
        'default' : {'fd_fcn' : lambda _: (np.zeros(1), np.zeros(1)), 'label_fcn' : lambda _: '', 'ylabel' : ''},
        'voltage' : {'fd_fcn' : solution.get_voltage, 'label_fcn' : lambda id: f'V({id})', 'ylabel' : 'V'},
        'current' : {'fd_fcn' : solution.get_current, 'label_fcn' : lambda id: f'I({id})', 'ylabel' : 'I'},
        'power' : {'fd_fcn' : solution.get_power, 'label_fcn' : lambda id: f'S({id})', 'ylabel' : 'S'}
    }
    fd_properties = fd_plot_config.get(type, fd_plot_config['default'])
    ylabel = ylabel if len(ylabel) > 0 else fd_properties['ylabel']

    @layout.legend(yoffset=1.2)
    @layout.grid()
    @layout.bode_like_plot(wmin=wmin, wmax=wmax, logw=logw, logy=logy, xlabel=xlabel, ylabel=ylabel)
    def plot_frequencydomain(fig:layout.Figure, ax:layout.Axes) -> layout.FigureAxes:
        ax[-2].set_ylim(bottom=abs_min)
        if abs_max > 0:
            ax[-2].set_ylim(top=abs_max)
        ax[-1].set_ylim(bottom=phi_min, top=phi_max)
        new_args = [functools.partial(a, fd_fcn=fd_properties['fd_fcn'], label_fcn=fd_properties['label_fcn']) for a in args]
        layout.apply_plt_fcn(fig, ax, *new_args)
        return fig, ax
    return plot_frequencydomain

def frequency_domain_plot(*args, layout_fcn:layout.Layout=layout.figure_two_rows, **kwargs) -> layout.FigureAxes:
    plot_frequencydomain = plot_frequencydomain_factory(*args, type='default', **kwargs)
    return plot_frequencydomain(*layout_fcn())

def discrete_frequencies_voltage_analysis(*args, layout_fcn:layout.Layout=layout.figure_two_rows, **kwargs) -> layout.FigureAxes:
    plot_frequencydomain = plot_frequencydomain_factory(*args, type='voltage', **kwargs)
    return plot_frequencydomain(*layout_fcn())

def discrete_frequencies_current_analysis(*args, layout_fcn:layout.Layout=layout.figure_two_rows, **kwargs) -> layout.FigureAxes:
    plot_frequencydomain = plot_frequencydomain_factory(*args, type='current', **kwargs)
    return plot_frequencydomain(*layout_fcn())

def discrete_frequencies_power_analysis(*args, layout_fcn:layout.Layout=layout.figure_two_rows, **kwargs) -> layout.FigureAxes:
    plot_frequencydomain = plot_frequencydomain_factory(*args, type='power', **kwargs)
    return plot_frequencydomain(*layout_fcn())


from typing import Mapping
from .node_analysis import state_space_matrices
from . import label_mapping as map
from ..network import Network
from .. import matrix_operations as mo
from ..matrix_operations import symbolic

class StateSpaceGenericOutput:
    def __init__(self, network: Network, c_values: Mapping[str, float | symbolic], l_values: Mapping[str, float | symbolic], matrix_ops: mo.MatrixOperations, node_index_mapper: map.NetworkMapper, voltage_source_index_mapper: map.SourceIndexMapper, current_source_index_mapper: map.SourceIndexMapper):
        self.network = network
        self.c_values = c_values
        self.l_values = l_values
        self.matrix_ops = matrix_ops
        self.A, self.B, self.C, self.D = state_space_matrices(
            network=self.network,
            c_values=self.c_values,
            l_values=self.l_values,
            matrix_ops=self.matrix_ops,
            node_mapper=node_index_mapper,
            voltage_source_mapper=voltage_source_index_mapper,
            current_source_mapper=current_source_index_mapper
        )
        self._node_label_mapping = node_index_mapper(self.network)
        self._voltage_source_label_mapping = voltage_source_index_mapper(self.network)
        self._current_source_label_mapping = current_source_index_mapper(self.network)

    def _row_for_potential(self, node_id: str, matrix: mo.Matrix) -> mo.Matrix:
        if node_id in self._node_label_mapping:
            return matrix[self._node_label_mapping[node_id]:self._node_label_mapping[node_id]+1, :] # type: ignore
        return self.matrix_ops.zeros((1, matrix.shape[1]))

    def c_row_for_potential(self, node_id: str) -> mo.Matrix:
        return self._row_for_potential(node_id, self.C)

    def c_row_voltage(self, branch_id: str) -> mo.Matrix:
        branch = self.network[branch_id]
        c_pos = self.c_row_for_potential(branch.node1)
        c_neg = self.c_row_for_potential(branch.node2)
        return c_pos - c_neg

    def c_row_current(self, branch_id: str) -> mo.Matrix:
        voltage_source_mapping = map.filter(self._voltage_source_label_mapping, lambda x: self.network[x].element.is_ideal_voltage_source)
        if branch_id in self.c_values:
            idx = list(self.c_values.keys()).index(branch_id)
            return self.c_values[branch_id] * self.A[idx, :] # type: ignore
        if branch_id in self._voltage_source_label_mapping:
            return self.C[voltage_source_mapping[branch_id] + self._node_label_mapping.N, :] # type: ignore
        if branch_id in self._current_source_label_mapping:
            return self.matrix_ops.zeros((1, self.C.shape[1]))
        branch = self.network[branch_id]
        c_pos = self.c_row_for_potential(branch.node1)
        c_neg = self.c_row_for_potential(branch.node2)
        return (c_pos - c_neg) / branch.element.Z

    def d_row_for_potential(self, node_id: str) -> mo.Matrix:
        return self._row_for_potential(node_id, self.D)

    def d_row_voltage(self, branch_id: str) -> mo.Matrix:
        branch = self.network[branch_id]
        d_pos = self.d_row_for_potential(branch.node1)
        d_neg = self.d_row_for_potential(branch.node2)
        return d_pos - d_neg

    def d_row_current(self, branch_id: str) -> mo.Matrix:
        if branch_id in self.c_values:
            idx = list(self.c_values.keys()).index(branch_id)
            return self.c_values[branch_id] * self.B[idx, :] # type: ignore
        if branch_id in self._voltage_source_label_mapping:
            return self.D[self._voltage_source_label_mapping[branch_id] + self._node_label_mapping.N, :] # type: ignore
        if branch_id in self._current_source_label_mapping:
            d_row = self.matrix_ops.zeros((1, self.D.shape[1]))
            d_row[0, self._current_source_label_mapping[branch_id]] = 1
            return d_row
        branch = self.network[branch_id]
        d_pos = self.d_row_for_potential(branch.node1)
        d_neg = self.d_row_for_potential(branch.node2)
        return (d_pos - d_neg) / branch.element.Z

    def extend_C_matrix(self, potential_nodes: list[str] = [], voltage_ids: list[str] = [], current_ids: list[str] = []) -> mo.Matrix:
        extended_C = self.matrix_ops.zeros((0, self.C.shape[1]))
        for id in potential_nodes:
            extended_C = self.matrix_ops.vstack((extended_C, self.c_row_for_potential(id)))
        for id in voltage_ids:
            extended_C = self.matrix_ops.vstack((extended_C, self.c_row_voltage(id)))
        for id in current_ids:
            extended_C = self.matrix_ops.vstack((extended_C, self.c_row_current(id)))
        return extended_C

    def extend_D_matrix(self, potential_nodes: list[str] = [], voltage_ids: list[str] = [], current_ids: list[str] = []) -> mo.Matrix:
        extended_D = self.matrix_ops.zeros((0, self.D.shape[1]))
        for id in potential_nodes:
            extended_D = self.matrix_ops.vstack((extended_D, self.d_row_for_potential(id)))
        for id in voltage_ids:
            extended_D = self.matrix_ops.vstack((extended_D, self.d_row_voltage(id)))
        for id in current_ids:
            extended_D = self.matrix_ops.vstack((extended_D, self.d_row_current(id)))
        return extended_D

    def sources(self) -> list[str]:
        current_sources = self._current_source_label_mapping.keys
        voltage_sources = [vs for vs in self._voltage_source_label_mapping.keys if vs not in self.l_values]
        return current_sources + voltage_sources

def numeric_state_space_model(network: Network, c_values: Mapping[str, float], l_values: Mapping[str, float], node_index_mapper: map.NetworkMapper = map.default_node_mapper, voltage_source_index_mapper: map.SourceIndexMapper = map.alphabetic_voltage_source_mapper, current_source_index_mapper: map.SourceIndexMapper = map.alphabetic_current_source_mapper) -> StateSpaceGenericOutput:
    return StateSpaceGenericOutput(
        network=network,
        c_values=c_values,
        l_values=l_values,
        matrix_ops=mo.NumPyMatrixOperations(),
        node_index_mapper=node_index_mapper,
        voltage_source_index_mapper=voltage_source_index_mapper,
        current_source_index_mapper=current_source_index_mapper
    )

def symbolic_state_space_model(network: Network, c_values: Mapping[str, symbolic], l_values: Mapping[str, symbolic], node_index_mapper: map.NetworkMapper = map.default_node_mapper, voltage_source_index_mapper: map.SourceIndexMapper = map.alphabetic_voltage_source_mapper, current_source_index_mapper: map.SourceIndexMapper = map.alphabetic_current_source_mapper) -> StateSpaceGenericOutput:
    return StateSpaceGenericOutput(
        network=network,
        c_values=c_values,
        l_values=l_values,
        matrix_ops=mo.SymPyMatrixOperations(),
        node_index_mapper=node_index_mapper,
        voltage_source_index_mapper=voltage_source_index_mapper,
        current_source_index_mapper=current_source_index_mapper
    )



from . import circuit as cc
from ..Network.NodalAnalysis import state_space_model as ssm
from ..SignalProcessing.state_space_model import NumericStateSpaceModel, SymbolicStateSpaceModel
from typing import Any

import numpy as np
import sympy as sp
from dataclasses import dataclass

@dataclass
class StateSpaceMatrixConstructor:

    _state_space_model: ssm.StateSpaceGenericOutput

    @property
    def A(self) -> Any:
        return self._state_space_model.A
    
    @property
    def B(self) -> Any:
        return self._state_space_model.B

    def C(self, potential_nodes: list[str], voltage_ids: list[str], current_ids: list[str]) -> Any:
        return self._state_space_model.extend_C_matrix(potential_nodes, voltage_ids, current_ids)

    def D(self, potential_nodes: list[str], voltage_ids: list[str], current_ids: list[str]) -> Any:
        return self._state_space_model.extend_D_matrix(potential_nodes, voltage_ids, current_ids)

    def c_d_row_for_potential(self, node_id: str) -> tuple[Any, Any]:
        return self._state_space_model.c_row_for_potential(node_id), self._state_space_model.d_row_for_potential(node_id)

    def c_d_row_for_voltage(self, node_id: str) -> tuple[Any, Any]:
        return self._state_space_model.c_row_voltage(node_id), self._state_space_model.d_row_voltage(node_id)

    def c_d_row_for_current(self, node_id: str) -> tuple[Any, Any]:
        return self._state_space_model.c_row_current(node_id), self._state_space_model.d_row_current(node_id)

    @property
    def sources(self) -> list[str]:
        return self._state_space_model.sources()

def numeric_state_space_model_constructor(circuit) -> StateSpaceMatrixConstructor:
    network = cc.transform_circuit(circuit, w=0)
    state_space_model = ssm.numeric_state_space_model(
        network=network,
        c_values={C.id : float(C.value['C']) for C in [c for c in circuit.components if c.type == 'capacitor']},
        l_values={L.id : float(L.value['L']) for L in [c for c in circuit.components if c.type == 'inductance']}
    )
    return StateSpaceMatrixConstructor(state_space_model)

def numeric_state_space_model(circuit: cc.Circuit, potential_nodes: list[str] = [], voltage_ids: list[str] = [], current_ids: list[str] = []) -> NumericStateSpaceModel:
    state_space_model = numeric_state_space_model_constructor(circuit)
    return NumericStateSpaceModel(
        A=np.array(state_space_model.A),
        B=np.array(state_space_model.B),
        C=np.array(state_space_model.C(potential_nodes, voltage_ids, current_ids)),
        D=np.array(state_space_model.D(potential_nodes, voltage_ids, current_ids))
    )

def symbolic_state_space_model_constructor(circuit) -> StateSpaceMatrixConstructor:
    def to_symbolic(id: str, value: str) -> sp.Symbol:
        sym_value = sp.sympify(value)
        if sym_value == sp.nan or id == value:
            return sp.Symbol(id, real=True, positive=True)
        return sym_value
    network = cc.transform_symbolic_circuit(circuit, s=sp.sympify(0))
    state_space_model = ssm.symbolic_state_space_model(
        network=network,
        c_values={C.id : to_symbolic(C.id, C.value['C']) for C in [c for c in circuit.components if c.type == 'capacitor']},
        l_values={L.id : to_symbolic(L.id, L.value['L']) for L in [c for c in circuit.components if c.type == 'inductance']}
    )
    return StateSpaceMatrixConstructor(state_space_model)

def symbolic_state_space_model(circuit: cc.Circuit, potential_nodes: list[str] = [], voltage_ids: list[str] = [], current_ids: list[str] = []) -> SymbolicStateSpaceModel:
    state_space_model = symbolic_state_space_model_constructor(circuit)
    return SymbolicStateSpaceModel(
        A=state_space_model.A,
        B=state_space_model.B,
        C=state_space_model.C(potential_nodes, voltage_ids, current_ids),
        D=state_space_model.D(potential_nodes, voltage_ids, current_ids)
    )


from .components import Component
import sympy as sp

def value_less_than_zero(value: str) -> bool:
    if len(value) == 0:
        return False
    x = sp.sympify(value)
    try:
        return (x < 0) == True
    except TypeError:
        return False

def resistor(id: str, nodes: tuple[str, str], R: str = '') -> Component:
    if value_less_than_zero(R):
        raise ValueError('R must be greater than zero.')
    return Component(
        type='resistor',
        id=id,
        value={'R': R if len(R) > 0 else id},
        nodes=nodes
        )

def conductance(id: str, nodes: tuple[str, str], G: str = '') -> Component:
    if value_less_than_zero(G):
        raise ValueError('G must be greater than zero.')
    return Component(
        type='conductance',
        id=id,
        value={'G': G if len(G) > 0 else id},
        nodes=nodes
        )

def capacitor(id: str, nodes: tuple[str, str], C: str = '') -> Component:
    if value_less_than_zero(C):
        raise ValueError('C must be greater than zero.')
    return Component(
        type='capacitor',
        id=id,
        value={'C': C if len(C) > 0 else id},
        nodes=nodes
    )

def inductance(id: str, nodes: tuple[str, str], L: str = '') -> Component:
    if value_less_than_zero(L):
        raise ValueError('L must be greater than zero.')
    return Component(
        type='inductance',
        id=id,
        value={'L': L if len(L) > 0 else id},
        nodes=nodes
    )

def impedance(id: str, nodes: tuple[str, str], Z: str = '') -> Component:
    return Component(
        type='impedance',
        id=id,
        value={'Z': Z if len(Z) > 0 else id},
        nodes=nodes
        )

def admittance(id: str, nodes: tuple[str, str], Y: str = '') -> Component:
    return Component(
        type='admittance',
        id=id,
        value={'Y': Y if len(Y) > 0 else id},
        nodes=nodes
        )

def voltage_source(id: str, nodes: tuple[str, str], V: str = '', R: str = '0') -> Component:
    return Component(
        type='dc_voltage_source',
        id=id,
        value={'V': V if len(V) > 0 else id, 'R': R, 'w': 0, 'phi': 0},
        nodes=nodes
        )

def current_source(id: str, nodes: tuple[str, str], I: str = '', G: str = '0') -> Component:
    return Component(
        type='dc_current_source',
        id=id,
        value={'I': I if len(I) > 0 else id, 'G': G, 'w': 0, 'phi': 0},
        nodes=nodes
        )

def short_circuit(id: str, nodes: tuple[str, str]) -> Component:
    return Component(
        type='short_circuit',
        id=id,
        nodes=nodes
    )        

def ground(id: str='gnd', nodes: tuple[str]=('0',)) -> Component:
    return Component(
        type='ground',
        id=id,
        nodes=nodes
    )


from ..Utils import ScientificFloat, ScientificComplex
from math import degrees, pi
from cmath import phase

blue = '#02468F'
red = '#D20000'
green = '#007355'

def print_complex(value: complex, unit: str = '', precision: int = 3, polar: bool = False, deg: bool = False) -> str:
    return str(ScientificComplex(
        value=value,
        unit=unit,
        precision=precision,
        polar=polar,
        use_exp_prefix=True,
        deg=deg,
        exp_prefixes={-6: 'u', -3: 'm', 3: 'k'},
        compact=True
    ))

def print_abs(value: complex, unit: str = '', precision: int = 3) -> str:
    return str(ScientificFloat(
        value=abs(value),
        unit=unit,
        precision=precision,
        use_exp_prefix=True,
        exp_prefixes={-6: 'u', -3: 'm', 3: 'k'},
    ))

def print_real(value: complex, unit: str = '', precision: int = 3) -> str:
    return str(ScientificFloat(
        value=value.real,
        unit=unit,
        precision=precision,
        use_exp_prefix=True,
        exp_prefixes={-6: 'u', -3: 'm', 3: 'k'},
    ))

def print_sinosoidal(value: complex, unit: str = '', precision: int = 3, w: float = 0, sin: bool = False, deg: bool = False, hertz: bool = False) -> str:
    abs_value = ScientificFloat(value=abs(value), unit=unit, precision=precision, use_exp_prefix=True, exp_prefixes={-6: 'u', -3: 'm', 3: 'k'})
    phase_value = phase(value)
    phase_value+= -pi/2 if sin else 0
    abs_phase_value = ScientificFloat(value=abs(degrees(phase_value)), unit='°', precision=precision) if deg else ScientificFloat(value=abs(phase_value), precision=precision)
    label = str(abs_value)
    if w == 0:
        return label
    label+= '·'
    label+= 'sin' if sin else 'cos'
    label+= '('
    label+= '2π·' if hertz else ''
    label+= str(ScientificFloat(w/2/pi, 'Hz', precision=precision, use_exp_prefix=True, exp_prefixes={-3: 'm', 3: 'k', 6: 'M', 9: 'G', 12: 'T'})) if hertz else str(ScientificFloat(w, '/s', precision=precision))
    label+= '·t'
    if abs(phase_value) > 1e-4:
        label+= '+' if phase_value > 0 else '-'
        label+= str(abs_phase_value)
    label+= ')'
    return label

def print_active_power(value: float, precision: int = 3) -> str:
    P = ScientificFloat(value=abs(value), unit='W', precision=precision, use_exp_prefix=True)
    P_sign = '↓' if value > 0 else '↑'
    label = f'{P}{P_sign}'
    return label

def print_active_reactive_power(value: complex, precision: int = 3) -> str:
    P = ScientificFloat(value=abs(value.real), unit='W', precision=precision, use_exp_prefix=True)
    Q = ScientificFloat(value=abs(value.imag), unit='var', precision=precision, use_exp_prefix=True)
    P_sign = '↓' if value.real > 0 else '↑'
    Q_sign = '↓' if value.imag > 0 else '↑'
    label = f'P: {P_sign}{P}'
    label+= f'\nQ: {Q_sign}{Q}' if Q.value > 1e-4 else ''
    return label

def print_resistance(R: float, precision: int = 3):
    return str(ScientificFloat(value=R, unit='Ω', use_exp_prefix=True, exp_prefixes={-3: 'm', 3: 'k', 6: 'M', 9: 'G'}, precision=precision))

def print_conductance(G: float, precision: int = 3):
    return str(ScientificFloat(value=G, unit='S', use_exp_prefix=True, exp_prefixes={-3: 'm', 3: 'k', 6: 'M', 9: 'G'}, precision=precision))

def print_impedance(Z: complex, precision: int = 3):
    return str(ScientificComplex(value=Z, unit='Ω', use_exp_prefix=True, exp_prefixes={-3: 'm', 3: 'k', 6: 'M', 9: 'G'}, precision=precision))

def print_capacitance(C: float, precision: int = 3):
    return str(ScientificFloat(value=C, unit='F', use_exp_prefix=True, exp_prefixes={-12: 'p', -9: 'n', -6: 'μ', -3: 'm'}, precision=precision))

def print_inductance(L: float, precision: int = 3):
    return str(ScientificFloat(value=L, unit='H', use_exp_prefix=True, exp_prefixes={-9: 'n', -6: 'μ', -3: 'm'}, precision=precision))


from .Elements import Schematic
from ..Circuit.dump_load import dictify_circuit
from .DiagramTranslator import circuit_translator
from . import Elements as simple_circuit_elements
import schemdraw.elements
import schemdraw.util
import schemdraw.transform
from typing import Any, TypedDict, Callable, TypeVar
from collections import ChainMap
from .. import dump_load
import functools

class SchemdrawObjectProperties(TypedDict):
    type: str
    values: dict[str, Any] | list

class SimpleCircuitObjectProperties(SchemdrawObjectProperties):
    name: str
    reverse: bool

T = TypeVar('T')

def schemdraw_object_properties(object: T, value_fcn: Callable[[T], dict[str, Any] | list]) -> SchemdrawObjectProperties:
    return SchemdrawObjectProperties(type=str(type(object).__name__), values=value_fcn(object))

def listify_point(p: schemdraw.util.Point) -> list[float]:
    return [serialize_schemdraw_element(p.x), serialize_schemdraw_element(p.y)]

def dictify_segment(s: schemdraw.segments.Segment) -> dict[str, Any]:
    return {
        'path' : serialize_schemdraw_element(s.path),
        'zorder' : serialize_schemdraw_element(s.zorder),
        'color' : serialize_schemdraw_element(s.color),
        'fill' : serialize_schemdraw_element(s.fill),
        'lw' : serialize_schemdraw_element(s.lw),
        'ls' : serialize_schemdraw_element(s.ls),
        'arrow' : serialize_schemdraw_element(s.arrow),
        'arrowwidth' : serialize_schemdraw_element(s.arrowwidth),
        'arrowlength' : serialize_schemdraw_element(s.arrowlength),
        'clip' : serialize_schemdraw_element(s.clip),
        'capstyle' : serialize_schemdraw_element(s.capstyle),
        'joinstyle' : serialize_schemdraw_element(s.joinstyle),
        'visible' : serialize_schemdraw_element(s.visible)
    }

def dictify_segment_text(s: schemdraw.segments.SegmentText) -> dict[str, Any]:
    return {
        'pos' : serialize_schemdraw_element(s.xy),
        'label' : serialize_schemdraw_element(s.text),
        'align' : serialize_schemdraw_element(s.align),
        'font' : serialize_schemdraw_element(s.font),
        'mathfont' : serialize_schemdraw_element(s.mathfont),
        'fontsize' : serialize_schemdraw_element(s.fontsize),
        'color' : serialize_schemdraw_element(s.color),
        'rotation' : serialize_schemdraw_element(s.rotation),
        'rotation_mode' : serialize_schemdraw_element(s.rotation_mode),
        'rotation_global' : serialize_schemdraw_element(s.rotation_global),
        'clip' : serialize_schemdraw_element(s.clip),
        'zorder' : serialize_schemdraw_element(s.zorder),
        'visible' : serialize_schemdraw_element(s.visible)
        }

def dictify_segment_circle(s: schemdraw.segments.SegmentCircle) -> dict[str, Any]:
    return {
        'center' : serialize_schemdraw_element(s.center),
        'radius' : serialize_schemdraw_element(s.radius),
        'color' : serialize_schemdraw_element(s.color),
        'lw' : serialize_schemdraw_element(s.lw),
        'ls' : serialize_schemdraw_element(s.ls),
        'fill' : serialize_schemdraw_element(s.fill),
        'clip' : serialize_schemdraw_element(s.clip),
        'zorder' : serialize_schemdraw_element(s.zorder),
        'ref' : serialize_schemdraw_element(s.endref),
        'visible' : serialize_schemdraw_element(s.visible)
        }

def dictify_transform(t: schemdraw.transform.Transform) -> dict[str, Any]:
    return {
        'theta' : serialize_schemdraw_element(t.theta),
        'globalshift' : serialize_schemdraw_element(t.shift),
        'localshift' : serialize_schemdraw_element(t.localshift),
        'zoom' : serialize_schemdraw_element(t.zoom)
    }

schemdraw_serializers = {
    str: lambda x: x,
    int: lambda x: x,
    float: lambda x: x,
    bool: lambda x: x,
    dict: lambda x: {k: serialize_schemdraw_element(v) for k, v in x.items()},
    list: lambda x: [serialize_schemdraw_element(e) for e in x],
    tuple: lambda x: [serialize_schemdraw_element(e) for e in x],
    ChainMap: lambda x: {k: serialize_schemdraw_element(v) for k, v in x.items()},
    schemdraw.util.Point: lambda x: schemdraw_object_properties(x, listify_point),
    schemdraw.segments.Segment: lambda x: schemdraw_object_properties(x, dictify_segment),
    schemdraw.segments.SegmentText: lambda x: schemdraw_object_properties(x, dictify_segment_text),
    schemdraw.segments.SegmentCircle: lambda x: schemdraw_object_properties(x, dictify_segment_circle),
    schemdraw.transform.Transform: lambda x: schemdraw_object_properties(x, dictify_transform)
}

def serialize_schemdraw_element(e):
    serialize = schemdraw_serializers.get(type(e), lambda _: None)
    return serialize(e)

def dictify_element(e: schemdraw.elements.Element) -> SimpleCircuitObjectProperties:
    return SimpleCircuitObjectProperties(
        type=e.type,
        name=e.name,
        reverse=e.is_reverse,
        values={
            '_userparams' : serialize_schemdraw_element(e._userparams),
            'segments' : serialize_schemdraw_element(e.segments),
            'params' : serialize_schemdraw_element(e.params),
            'anchors' : serialize_schemdraw_element(e.anchors),
            'absanchors' : serialize_schemdraw_element(e.absanchors),
            'transform' : serialize_schemdraw_element(e.transform),
            'absdrop' : serialize_schemdraw_element(e.absdrop)
        }
    )

def schematic_to_dict(d: schemdraw.Drawing) -> list[SimpleCircuitObjectProperties]:
    return [dictify_element(e) for e in d.elements]

schemdraw_deserializers = {
    str(schemdraw.segments.Segment.__name__) : lambda x: schemdraw.segments.Segment(**deserialize_schemdraw_elements(x)), # type: ignore
    str(schemdraw.segments.SegmentText.__name__) : lambda x: schemdraw.segments.SegmentText(**deserialize_schemdraw_elements(x)),
    str(schemdraw.segments.SegmentCircle.__name__) : lambda x: schemdraw.segments.SegmentCircle(**deserialize_schemdraw_elements(x)),
    str(schemdraw.util.Point.__name__) : lambda *x: schemdraw.util.Point(*x),
    str(schemdraw.transform.Transform.__name__) : lambda x: schemdraw.transform.Transform(**deserialize_schemdraw_elements(x)),
}

def deserialize_schemdraw_elements(element: dict[str, Any] | list) -> Any:
    if type(element) == list:
        return [deserialize_schemdraw_elements(e) for e in element]
    if type(element) == dict:
        try:
            return schemdraw_deserializers[element['type']](element['values'])
        except KeyError:
            return {k: deserialize_schemdraw_elements(v) for k, v in element.items()}
    return element

def combine_to_complex(real_imag: tuple[str, str], z: str, kv: dict[str, Any]) -> dict[str, Any]:
    kv.update({z: complex(kv.pop(real_imag[0], 0), kv.pop(real_imag[1], 0))})
    return kv

simple_circuit_element_types = {
    'voltage_source' : lambda **kwargs: simple_circuit_elements.VoltageSource(**kwargs),
    'current_source' : lambda **kwargs: simple_circuit_elements.CurrentSource(**kwargs),
    'ac_voltage_source' : lambda **kwargs: simple_circuit_elements.ACVoltageSource(**kwargs),
    'ac_current_source' : lambda **kwargs: simple_circuit_elements.ACCurrentSource(**kwargs),
    'rect_voltage_source' : lambda **kwargs: simple_circuit_elements.RectVoltageSource(**kwargs),
    'rect_current_source' : lambda **kwargs: simple_circuit_elements.RectCurrentSource(**kwargs),
    'complex_voltage_source' : lambda **kwargs: simple_circuit_elements.ComplexVoltageSource(**combine_to_complex(('V_real', 'V_imag'), 'V', kwargs)),
    'complex_current_source' : lambda **kwargs: simple_circuit_elements.ComplexCurrentSource(**combine_to_complex(('I_real', 'I_imag'), 'I', kwargs)), 
    'resistor' : lambda **kwargs: simple_circuit_elements.Resistor(**kwargs),
    'conductance' : lambda **kwargs: simple_circuit_elements.Conductance(**kwargs),
    'impedance' : lambda **kwargs: simple_circuit_elements.Impedance(**combine_to_complex(('R', 'X'), 'Z', kwargs)),
    'admittance' : lambda **kwargs: simple_circuit_elements.Admittance(**combine_to_complex(('G', 'B'), 'Y', kwargs)),
    'capacitor' : lambda **kwargs: simple_circuit_elements.Capacitor(**kwargs),
    'inductance' : lambda **kwargs: simple_circuit_elements.Inductance(**kwargs),
    'ground' : lambda **kwargs: simple_circuit_elements.Ground(**kwargs),
    'line' : lambda **kwargs: simple_circuit_elements.Line(**kwargs)
}

def undictify_element(element_dict: dict[str, Any], circuit_dict: dict[str, Any]) -> schemdraw.elements.Element:
    kwargs = deserialize_schemdraw_elements(element_dict['values']['_userparams'])
    kwargs.update({'name': element_dict.get('name', '')})
    kwargs.update({'reverse': element_dict.get('reverse', False)})
    if element_dict['name'] in circuit_dict.keys():
        kwargs.update(circuit_dict[element_dict['name']])
    try:
        element = simple_circuit_element_types[element_dict['type']](**kwargs)
    except KeyError:
        element = simple_circuit_elements.Element(**kwargs)
    element.segments=deserialize_schemdraw_elements(element_dict['values']['segments'])
    element.params=deserialize_schemdraw_elements(element_dict['values']['params'])
    element.anchors=deserialize_schemdraw_elements(element_dict['values']['anchors'])
    element.absanchors=deserialize_schemdraw_elements(element_dict['values']['absanchors'])
    element.transform=deserialize_schemdraw_elements(element_dict['values']['transform'])
    element.absdrop=deserialize_schemdraw_elements(element_dict['values']['absdrop'])
    return element

def undictify_schematic(schematic_dict: dict) -> Schematic:
    schematic = Schematic()
    circuit_dict = {c['id']: c['value'] for c in schematic_dict['circuit']['components']}
    schematic.elements.extend([undictify_element(e, circuit_dict) for e in schematic_dict['simple_circuit']])
    return schematic

def dictify_all(schematic: Schematic) -> dict:
    return {
        'circuit': dictify_circuit(circuit_translator(schematic)),
        'simple_circuit': schematic_to_dict(schematic)
    }

serialize = functools.partial(dump_load.serialize, dict_processor=dictify_all)
dump = functools.partial(dump_load.dump, dump_fcn=serialize)

deserialize = functools.partial(dump_load.deserialize, dict_preprocessor=undictify_schematic)
load = functools.partial(dump_load.load, deserialize_fcn=deserialize)


import schemdraw.elements
from .elements import InductorDIN, SourceIDIN2
from .elements import SourceUDIN
from .elements import SourceIDIN

STYLE_DIN = schemdraw.elements.STYLE_IEC.copy()
STYLE_DIN.update({
    'Inductor': InductorDIN,
    'SourceV': SourceUDIN,
    'SourceI': SourceIDIN,
    'SourceIArrow': SourceIDIN2,
    })



from CircuitCalculator.Circuit.circuit import AmbiguousComponentID
from CircuitCalculator.dump_load import FormatError

class UnknownCircuitElement(Exception):
    def __init__(self, element_type: str) -> None:
        super().__init__(f'Unknown circuit element type "{element_type}".')
        self.element_type = element_type

class MissingArgument(Exception):
    def __init__(self, missing_argument: str, provided_arguments: str) -> None:
        super().__init__(f'Missing argument: {missing_argument}. Given arguments: {provided_arguments}.')
        self.missing = missing_argument
        self.provided_arguments = provided_arguments

class UnknownArgument(Exception):
    def __init__(self, unknown_argument: str, method: str = '') -> None:
        super().__init__(f"Unknown argument {unknown_argument}" + f" for method '{method}'." if len(method) > 0 else '.')
        self.unknown_argument = unknown_argument
        self.method = method

class UnknownSolutionType(Exception):
    def __init__(self, solution_type: str, available_solutions: list[str]) -> None:
        super().__init__(f'Unknown solution type "{solution_type}". Available solutions: {", ".join(available_solutions)}.')
        self.solution_type = solution_type
        self.available_solutions = available_solutions

class SolutionUsageError(Exception):
    def __init__(self, solution_type: str, unknown_arguments: list[str]) -> None:
        error_message = f'Cannot use solution "{solution_type}" with arguments: {", ".join(unknown_arguments)}.'
        if len(unknown_arguments) == 0:
            error_message = f'Cannot use solution "{solution_type}".'
        if len(unknown_arguments) == 1:
            error_message = f'Cannot use solution "{solution_type}" with argument: {unknown_arguments[0]}.'
        super().__init__(error_message)
        self.solution_type = solution_type
        self.unknown_arguments = unknown_arguments

class IllegalElementValue(Exception): ...

class EmptyCircuit(Exception): ...

simulation_exceptions : tuple[type[Exception], ...] = (
    UnknownCircuitElement,
    MissingArgument,
    IllegalElementValue,
    UnknownArgument,
    AmbiguousComponentID,
    FileNotFoundError,
    FormatError,
    UnknownSolutionType,
    SolutionUsageError,
    EmptyCircuit,
    KeyError
)


from typing import Optional
from .schematic import draw_schematic, create_schematic
from . import errors
from matplotlib.axes import Axes
from CircuitCalculator.dump_load import load, FormatError
import CircuitCalculator.Circuit.solution as solution
from CircuitCalculator.Circuit.circuit import Circuit
from CircuitCalculator.SimpleCircuit.DiagramTranslator import circuit_translator

def load_simulation_file(name: str) -> dict:
    try:
        return load(name)
    except FileNotFoundError as e:
        raise FileNotFoundError(f'Simulation file "{name}" does not exist.') from e
    except FormatError as e:
        raise FormatError(f'Cannot parse "{name}" as simulation file due to format issues.') from e
    
def show_schematic(data: dict, ax: Optional[Axes] = None) -> None:
    try:
        draw_schematic(data, ax)
    except errors.simulation_exceptions as e:
        print(e)
        return

def show_schematic_from_file(name: str, ax: Optional[Axes] = None) -> None:
    try:
        data = load_simulation_file(name)
        show_schematic(data, ax)
    except errors.simulation_exceptions as e:
        print(e)
        return

solutions : dict[str, type[solution.CircuitSolution]]= {
    'dc': solution.DCSolution,
    'complex': solution.ComplexSolution,
    'time_domain': solution.TimeDomainSolution,
    'frequency_domain': solution.FrequencyDomainSolution,
    'transient': solution.TransientSolution,
    'symbolic': solution.SymbolicSolution
}

def get_solution(solution_type: str, circuit: Circuit, **kwargs) -> solution.CircuitSolution:
    try:
        solution = solutions[solution_type]
    except KeyError as e:
        raise errors.UnknownSolutionType(solution_type, list(solutions.keys())) from e
    try:
        return solution(circuit, **kwargs)
    except TypeError as e:
        raise errors.SolutionUsageError(solution_type, list(kwargs.keys())) from e

def simulate_schematic(data: dict, solution_type: str, **kwargs) -> solution.CircuitSolution:
    try:
        schematic = create_schematic(data)
        circuit = circuit_translator(schematic)
        return get_solution(solution_type, circuit, **kwargs)
    except errors.simulation_exceptions as e:
        print(e)
        return solution.EmptySolution()

def simulate_schematic_from_file(name: str, solution_type: str, **kwargs) -> solution.CircuitSolution:
    try:
        data = load_simulation_file(name)
        return simulate_schematic(data, solution_type, **kwargs)
    except errors.simulation_exceptions as e:
        print(e)
        return solution.EmptySolution()


from . import Elements as elm
from ..Circuit.Components import components as cp
from .SchemdrawTranslatorTypes import ElementTranslatorMap
from math import pi, inf


def resistor_translator(element: elm.Resistor, nodes: tuple[str, ...]) -> cp.Component:
    return cp.resistor(id=element.name, nodes=(nodes[0], nodes[1]), R=element.R)

def impedance_translator(element: elm.Impedance, nodes: tuple[str, ...]) -> cp.Component:
    return cp.impedance(id=element.name, nodes=(nodes[0], nodes[1]), Z=element.Z)

def conductance_translator(element: elm.Conductance, nodes: tuple[str, ...]) -> cp.Component:
    return cp.conductance(nodes=(nodes[0], nodes[1]), id=element.name, G=element.G)

def capacitor_translator(element: elm.Capacitor, nodes: tuple[str, ...]) -> cp.Component:
    return cp.capacitor(nodes=(nodes[0], nodes[1]), id=element.name, C=element.C)

def inductance_translator(element: elm.Inductance, nodes: tuple[str, ...]) -> cp.Component:
    return cp.inductance(nodes=(nodes[0], nodes[1]), id=element.name, L=element.L)

def lamp_translator(element: elm.Lamp, nodes: tuple[str, ...]) -> cp.Component:
    return cp.lamp(nodes=(nodes[0], nodes[1]), id=element.name, P=element.P_ref, V_ref=element.V_ref)

def ground_translator(element: elm.Ground, nodes: tuple[str, ...]) -> cp.Component:
    return cp.ground(nodes=(nodes[0],), id=element.name)

def dc_voltage_source_translator(element: elm.VoltageSource, nodes: tuple[str, ...]) -> cp.Component:
    return cp.dc_voltage_source(
        nodes=(nodes[0], nodes[1]) if not element.is_reverse else (nodes[1], nodes[0]),
        id=element.name,
        V=element.V.real if not element.is_reverse else -element.V.real
    )

def complex_voltage_source_translator(element: elm.ComplexVoltageSource, nodes: tuple[str, ...]) -> cp.Component:
    return cp.complex_voltage_source(
        nodes=(nodes[0], nodes[1]) if not element.is_reverse else (nodes[1], nodes[0]),
        id=element.name,
        V=element.V if not element.is_reverse else -element.V
    )

def dc_current_source_translator(element: elm.CurrentSource, nodes: tuple[str, ...]) -> cp.Component:
    return cp.dc_current_source(
        nodes=(nodes[0], nodes[1]) if not element.is_reverse else (nodes[1], nodes[0]),
        id=element.name,
        I=element.I.real if not element.is_reverse else -element.I.real
    )

def complex_current_source_translator(element: elm.ComplexCurrentSource, nodes: tuple[str, ...]) -> cp.Component:
    return cp.complex_current_source(
        nodes=(nodes[0], nodes[1]) if not element.is_reverse else (nodes[1], nodes[0]),
        id=element.name,
        I=element.V if not element.is_reverse else -element.I
    )

def ac_voltage_source_translator(element: elm.ACVoltageSource, nodes: tuple[str, ...]) -> cp.Component:
    return cp.ac_voltage_source(
        nodes=(nodes[0], nodes[1]) if not element.is_reverse else (nodes[1], nodes[0]),
        id=element.name,
        V=element.V if not element.is_reverse else -element.V,
        w=element.w,
        phi=element.phi*pi/180 if element.deg else element.phi
    )

def ac_current_source_translator(element: elm.ACCurrentSource, nodes: tuple[str, ...]) -> cp.Component:
    return cp.ac_current_source(
        nodes=(nodes[0], nodes[1]) if not element.is_reverse else (nodes[1], nodes[0]),
        id=element.name,
        I=element.I if not element.is_reverse else -element.I,
        w=element.w,
        phi=element.phi*pi/180 if element.deg else element.phi
    )

def rect_voltage_source_translator(element: elm.RectVoltageSource, nodes: tuple[str, ...]) -> cp.Component:
    from ..SignalProcessing.periodic_functions import RectFunction
    return cp.periodic_voltage_source(
        nodes=(nodes[0], nodes[1]) if not element.is_reverse else (nodes[1], nodes[0]),
        id=element.name,
        wavetype=RectFunction.wavetype,
        V=element.V if not element.is_reverse else -element.V,
        w=element.w,
        phi=element.phi*pi/180 if element.deg else element.phi
    )

def rect_current_source_translator(element: elm.RectCurrentSource, nodes: tuple[str, ...]) -> cp.Component:
    from ..SignalProcessing.periodic_functions import RectFunction
    return cp.periodic_current_source(
        nodes=(nodes[0], nodes[1]) if not element.is_reverse else (nodes[1], nodes[0]),
        id=element.name,
        wavetype=RectFunction.wavetype,
        I=element.I if not element.is_reverse else -element.I,
        w=element.w,
        phi=element.phi*pi/180 if element.deg else element.phi
    )

def tri_voltage_source_translator(element: elm.TriangleVoltageSource, nodes: tuple[str, ...]) -> cp.Component:
    from ..SignalProcessing.periodic_functions import TriFunction
    return cp.periodic_voltage_source(
        nodes=(nodes[0], nodes[1]) if not element.is_reverse else (nodes[1], nodes[0]),
        id=element.name,
        wavetype=TriFunction.wavetype,
        V=element.V if not element.is_reverse else -element.V,
        w=element.w,
        phi=element.phi*pi/180 if element.deg else element.phi
    )

def tri_current_source_translator(element: elm.TriangleCurrentSource, nodes: tuple[str, ...]) -> cp.Component:
    from ..SignalProcessing.periodic_functions import TriFunction
    return cp.periodic_current_source(
        nodes=(nodes[0], nodes[1]) if not element.is_reverse else (nodes[1], nodes[0]),
        id=element.name,
        wavetype=TriFunction.wavetype,
        I=element.I if not element.is_reverse else -element.I,
        w=element.w,
        phi=element.phi*pi/180 if element.deg else element.phi
    )

def saw_voltage_source_translator(element: elm.SawtoothVoltageSource, nodes: tuple[str, ...]) -> cp.Component:
    from ..SignalProcessing.periodic_functions import SawFunction
    return cp.periodic_voltage_source(
        nodes=(nodes[0], nodes[1]) if not element.is_reverse else (nodes[1], nodes[0]),
        id=element.name,
        wavetype=SawFunction.wavetype,
        V=element.V if not element.is_reverse else -element.V,
        w=element.w,
        phi=element.phi*pi/180 if element.deg else element.phi
    )

def saw_current_source_translator(element: elm.SawtoothCurrentSource, nodes: tuple[str, ...]) -> cp.Component:
    from ..SignalProcessing.periodic_functions import SawFunction
    return cp.periodic_current_source(
        nodes=(nodes[0], nodes[1]) if not element.is_reverse else (nodes[1], nodes[0]),
        id=element.name,
        wavetype=SawFunction.wavetype,
        I=element.I if not element.is_reverse else -element.I,
        w=element.w,
        phi=element.phi*pi/180 if element.deg else element.phi
    )

def linear_current_source_translator(element: elm.RealCurrentSource, nodes: tuple[str, ...]) -> cp.Component:
    return cp.dc_current_source(
        nodes=(nodes[0], nodes[1]) if not element.is_reverse else (nodes[1], nodes[0]),
        id=element.name,
        I=element.I.real if not element.is_reverse else -element.I.real,
        G=element.G
    )

def linear_voltage_source_translator(element: elm.RealVoltageSource, nodes: tuple[str, ...]) -> cp.Component:
    return cp.dc_voltage_source(
        nodes=(nodes[0], nodes[1]) if not element.is_reverse else (nodes[1], nodes[0]),
        id=element.name,
        V=element.V.real if not element.is_reverse else -element.V.real,
        R=element.R
    )

def short_circuit_translator(element: elm.LabeledLine, nodes: tuple[str, ...]) -> cp.Component:
    return cp.short_circuit(
        nodes=(nodes[0], nodes[1]) if not element.is_reverse else (nodes[1], nodes[0]),
        id=element.name,
    )

def switch_translator(element: elm.Switch, nodes: tuple[str, ...]) -> cp.Component | None:
    if element.state == element.state.OPEN:
        return cp.resistor(nodes=(nodes[0], nodes[1]), id=element.name, R=inf)
    return cp.resistor(nodes=(nodes[0], nodes[1]), id=element.name, R=1e-12)

def none_translator(*_) -> None:
    return None

circuit_translator_map : ElementTranslatorMap = {
    elm.Resistor : resistor_translator,
    elm.Impedance : impedance_translator,
    elm.Conductance : conductance_translator,
    elm.VoltageSource : dc_voltage_source_translator,
    elm.ComplexVoltageSource : complex_voltage_source_translator,
    elm.CurrentSource : dc_current_source_translator,
    elm.ComplexCurrentSource : complex_current_source_translator,
    elm.ACVoltageSource : ac_voltage_source_translator,
    elm.ACCurrentSource : ac_current_source_translator,
    elm.RectVoltageSource : rect_voltage_source_translator,
    elm.RectCurrentSource : rect_current_source_translator,
    elm.TriangleVoltageSource : tri_voltage_source_translator,
    elm.TriangleCurrentSource : tri_current_source_translator,
    elm.SawtoothVoltageSource : saw_voltage_source_translator,
    elm.SawtoothCurrentSource : saw_current_source_translator,
    elm.Capacitor : capacitor_translator,
    elm.Inductance : inductance_translator,
    elm.Lamp : lamp_translator,
    elm.Ground : ground_translator,
    elm.Line: none_translator,
    elm.LabeledLine: short_circuit_translator,
    elm.Node: none_translator,
    elm.LabelNode: none_translator,
    elm.RealCurrentSource: linear_current_source_translator,
    elm.RealVoltageSource: linear_voltage_source_translator,
    elm.Switch: switch_translator,
    elm.VoltageLabel: none_translator,
    elm.CurrentLabel: none_translator,
    elm.PowerLabel: none_translator,
    elm.Element: none_translator,
}



import sympy as sp
from ..Network import symbolic_elements as elm
from ..Network import network as ntw
from typing import Callable, TypeVar
from .Components import components as cp

CircuitComponent = TypeVar("CircuitComponent", bound=cp.Component)
CircuitComponentTranslator = Callable[[cp.Component, sp.Symbol], ntw.Branch]

def resistor(resistor: cp.Component, _: sp.Symbol) -> ntw.Branch:
    R = sp.sympify(resistor.value.get('R', 'nan'))
    if R == sp.nan or resistor.id == resistor.value['R']:
        R = sp.Symbol(resistor.id, real=True, positive=True)
    return ntw.Branch(resistor.nodes[0], resistor.nodes[1], elm.resistor(resistor.id, R))

def impedance(impedance: cp.Component, _: sp.Symbol) -> ntw.Branch:
    Z = sp.sympify(impedance.value.get('Z', 'nan'))
    if Z == sp.nan or impedance.id == impedance.value['Z']:
        Z = sp.Symbol(impedance.id, complex=True)
    return ntw.Branch(impedance.nodes[0], impedance.nodes[1], elm.impedance(impedance.id, Z))

def capacitor(capacitor: cp.Component, s: sp.Symbol) -> ntw.Branch:
    C = sp.sympify(capacitor.value['C'])
    if C == sp.nan or capacitor.id == capacitor.value['C']:
        C = sp.Symbol(capacitor.id, real=True, positive=True)
    return ntw.Branch(
        capacitor.nodes[0],
        capacitor.nodes[1],
        elm.admittance(capacitor.id, s*C)  # type: ignore
    )

def inductance(inductance: cp.Component, s: sp.Symbol) -> ntw.Branch:
    L = sp.sympify(inductance.value['L'])
    if L == sp.nan or inductance.id == inductance.value['L']:
        L = sp.Symbol(inductance.id, real=True, positive=True)
    return ntw.Branch(
        inductance.nodes[0],
        inductance.nodes[1],
        elm.impedance(inductance.id, s*L) # type: ignore
    )

def voltage_source(voltage_source: cp.Component, _: sp.Symbol) -> ntw.Branch:
    V = sp.sympify(voltage_source.value.get('V', 'nan'))
    Z = sp.sympify(voltage_source.value.get('R', 0))
    if V == sp.nan:
        V = sp.Symbol(voltage_source.id)
    element = elm.voltage_source(voltage_source.id, V, Z)
    return ntw.Branch(
        voltage_source.nodes[0],
        voltage_source.nodes[1],
        element)

def current_source(current_source: cp.Component, _: sp.Symbol) -> ntw.Branch:
    cs_I = sp.sympify(current_source.value.get('I', 'nan'))
    cs_G = sp.sympify(current_source.value.get('G', 0))
    if cs_I == sp.nan:
        cs_I = sp.Symbol(current_source.id)
    element = elm.current_source(current_source.id, cs_I, cs_G)
    return ntw.Branch(
        current_source.nodes[0],
        current_source.nodes[1],
        element
    )

def short_circuit(short_circuit: cp.Component, _: sp.Symbol) -> ntw.Branch:
    return ntw.Branch(
        short_circuit.nodes[0],
        short_circuit.nodes[1],
        elm.short_circuit(short_circuit.id)
    )

transformers : dict[str, CircuitComponentTranslator] = {
    'resistor' : resistor,
    'impedance' : impedance,
    'capacitor' : capacitor,
    'inductance' : inductance,
    'voltage_source' : voltage_source,
    'current_source' : current_source,
    'short_circuit' : short_circuit,
    'dc_voltage_source' : voltage_source,
    'dc_current_source' : current_source
}


import numpy as np
import itertools
from typing import Any, Callable, Mapping
from ..network import Network
from .. import matrix_operations as mo
from ..matrix_operations import symbolic
from .. import transformers as trf
from . import label_mapping as map

class DimensionError(Exception):
    ...

def admittance_connected_to(network: Network, node: str, me: Callable[[Any], mo.MatrixElement]) -> complex:
    return sum([e.value for e in [me(b.element.Y) for b in network.branches_connected_to(node)] if e.isfinite])

def admittance_between(network: Network, node1: str, node2: str, me: Callable[[Any], mo.MatrixElement]) -> complex:
    return sum([e.value for e in [me(b.element.Y) for b in network.branches_between(node1, node2)] if e.isfinite])

def connected_nodes(network: Network, node: str) -> list[str]:
    return [b.node1 if b.node1 != node else b.node2 for b in network.branches_connected_to(node)]

def node_admittance_matrix(network: Network, matrix_ops: mo.MatrixOperations = mo.NumPyMatrixOperations(), node_index_mapper: map.NetworkMapper = map.default_node_mapper) -> mo.Matrix:
    def node_matrix_element(i_label:str, j_label:str) -> complex:
        if i_label == j_label:
            return admittance_connected_to(no_voltage_sources_network, i_label, matrix_ops.elm)
        return -admittance_between(no_voltage_sources_network, i_label, j_label, matrix_ops.elm)
    node_mapping = node_index_mapper(network)
    no_voltage_sources_network = trf.remove_elements(network, [b.id for b in network.branches if b.element.is_ideal_voltage_source])
    Y = matrix_ops.zeros((node_mapping.N, node_mapping.N))
    for i_label, j_label in itertools.product(node_mapping, repeat=2):
        Y[node_mapping(i_label, j_label)] = node_matrix_element(i_label, j_label)
    return Y

def voltage_source_incidence_matrix(network: Network, matrix_ops: mo.MatrixOperations = mo.NumPyMatrixOperations(), node_mapper: map.NetworkMapper = map.default_node_mapper, source_mapper: map.SourceIndexMapper = map.alphabetic_voltage_source_mapper) -> mo.Matrix:
    def voltage_source_direction(voltage_source: str, node: str) -> int:
        if network[voltage_source].node1 == node:
            return 1
        if network[voltage_source].node2 == node:
            return -1
        return 0
    node_index = node_mapper(network)
    vs_index = source_mapper(network)
    A = matrix_ops.zeros((node_index.N, vs_index.N))
    for node, vs in itertools.product(node_index.keys, vs_index.keys):
        A[node_index[node], vs_index[vs]] = voltage_source_direction(vs, node)
    return A

def nodal_analysis_coefficient_matrix(network: Network, matrix_ops: mo.MatrixOperations = mo.NumPyMatrixOperations(), node_mapper: map.NetworkMapper = map.default_node_mapper, source_mapper: map.SourceIndexMapper = map.alphabetic_voltage_source_mapper) -> mo.Matrix:
    Y = node_admittance_matrix(network, matrix_ops, node_mapper)
    B = voltage_source_incidence_matrix(network, matrix_ops, node_mapper, source_mapper)
    Z = matrix_ops.zeros((B.shape[1], B.shape[1]))
    return matrix_ops.vstack((matrix_ops.hstack((Y, B)), matrix_ops.hstack((B.T, Z))))

def source_incidence_matrix(network: Network, node_mapper: map.NetworkMapper = map.default_node_mapper, source_mapper: map.SourceIndexMapper = map.alphabetic_current_source_mapper) -> np.ndarray:
    node_index = node_mapper(network)
    cs_index = source_mapper(network)
    Q = np.zeros((node_index.N, cs_index.N))
    for cs in cs_index.keys:
        source_element = network[cs]
        if network.node_zero_label != source_element.node1:
            Q[node_index[source_element.node1]][cs_index[cs]] = -1
        if network.node_zero_label != network[cs].node2:
            Q[node_index[source_element.node2]][cs_index[cs]] = 1
    return Q

def current_source_vector(network: Network, matrix_ops: mo.MatrixOperations = mo.NumPyMatrixOperations(), source_mapper: map.SourceIndexMapper = map.alphabetic_current_source_mapper) -> mo.Matrix:
    cs_index = map.filter(source_mapper(network), lambda x: network[x].element.is_current_source)
    return matrix_ops.column_vector([network[x].element.I for x in cs_index.keys])

def current_source_incidence_vector(network: Network, matrix_ops: mo.MatrixOperations = mo.NumPyMatrixOperations(), node_mapper: map.NetworkMapper = map.default_node_mapper, source_mapper: map.SourceIndexMapper = map.alphabetic_current_source_mapper) -> np.ndarray:
    Q = source_incidence_matrix(network, node_mapper=node_mapper, source_mapper=source_mapper)
    Is = current_source_vector(network, matrix_ops=matrix_ops, source_mapper=source_mapper)
    return Q@Is

def nodal_analysis_constants_vector(network: Network, matrix_ops: mo.MatrixOperations = mo.NumPyMatrixOperations(), node_mapper: map.NetworkMapper = map.default_node_mapper, current_source_mapper: map.SourceIndexMapper = map.alphabetic_current_source_mapper, voltage_source_mapper: map.SourceIndexMapper = map.alphabetic_voltage_source_mapper) -> mo.Matrix:
    I = current_source_incidence_vector(network, matrix_ops, node_mapper, current_source_mapper)

    vs_mapping = voltage_source_mapper(network)
    V = matrix_ops.column_vector([network[vs].element.V for vs in vs_mapping.keys])
    return matrix_ops.vstack((I, V))

def open_circuit_impedance(network: Network, node1: str, node2: str, matrix_ops: mo.MatrixOperations = mo.NumPyMatrixOperations(), node_index_mapper: map.NetworkMapper = map.default_node_mapper) -> symbolic:
    if node1 == node2:
        return matrix_ops.elm(0).value
    if any([b.element.is_ideal_voltage_source for b in network.branches_between(node1, node2)]):
        return matrix_ops.elm(0).value
    if network.is_zero_node(node1):
        node1, node2 = node2, node1
    network = trf.switch_ground_node(network=network, new_ground=node2)
    Y = nodal_analysis_coefficient_matrix(network, matrix_ops=matrix_ops, node_mapper=node_index_mapper)
    Y = matrix_ops.delete(Y, np.where(~matrix_ops.any_element(Y, axis=0))[0].tolist(), axis=1)
    Y = matrix_ops.delete(Y, np.where(~matrix_ops.any_element(Y, axis=1))[0].tolist(), axis=0)
    Z = matrix_ops.inv(Y)
    i1 = node_index_mapper(network)[node1]
    return Z.diagonal()[i1]

def element_impedance(network: Network, element: str, matrix_ops: mo.MatrixOperations = mo.NumPyMatrixOperations(), node_index_mapper: map.NetworkMapper = map.default_node_mapper) -> symbolic:
    return open_circuit_impedance(
        network=trf.remove_element(network, element),
        node1=network[element].node1,
        node2=network[element].node2,
        matrix_ops=matrix_ops,
        node_index_mapper=node_index_mapper
    )

def state_space_matrices(network: Network, c_values: Mapping[str, float | symbolic] = {}, l_values: Mapping[str, float | symbolic] = {}, matrix_ops: mo.MatrixOperations = mo.NumPyMatrixOperations(), node_mapper: map.NetworkMapper = map.default_node_mapper, current_source_mapper: map.SourceIndexMapper = map.alphabetic_current_source_mapper, voltage_source_mapper: map.SourceIndexMapper = map.alphabetic_voltage_source_mapper) -> tuple[mo.Matrix, mo.Matrix, mo.Matrix, mo.Matrix]:
    def element_incidence_matrix(values: Mapping[str, float | symbolic]) -> np.ndarray:
        node_mapping = node_mapper(network)
        Delta = np.zeros((len(values), node_mapping.N))
        for (k, value), (i_label) in itertools.product(enumerate(values), node_mapping):
            if i_label == network[value].node1:
                Delta[k][node_mapping(i_label)] = +1
            if i_label == network[value].node2:
                Delta[k][node_mapping(i_label)] = -1
        return np.hstack((Delta, np.zeros((Delta.shape[0], voltage_source_mapper(network).N))))
    def source_and_inductance_incidence_matrix(values: Mapping[str, float | symbolic]) -> tuple[np.ndarray, np.ndarray]:
        voltage_source_mapping_all = voltage_source_mapper(network)
        source_mapping_all = map.default_source_mapper(network)
        Qi = source_incidence_matrix(network=network, node_mapper=node_mapper, source_mapper=current_source_mapper)
        Q = np.zeros((voltage_source_mapping_all.N, voltage_source_mapping_all.N), dtype=int)
        for i in voltage_source_mapping_all.values:
            Q[i][i] = 1
        Q = np.vstack((np.hstack( (Qi, np.zeros((Qi.shape[0], Q.shape[1]) ))),
                    np.hstack( (np.zeros((Q.shape[0], Qi.shape[1])), Q) )))
        QS = Q[:,[source_mapping_all[l] for l in source_mapping_all if l not in values]]
        QL = Q[:,[source_mapping_all[l] for l in source_mapping_all if l in values]]
        return QS, QL
    def value_matrix(c_values: Mapping[str, float | symbolic], l_values: Mapping[str, float | symbolic]) -> np.ndarray:
        return matrix_ops.vstack((
            matrix_ops.hstack(( matrix_ops.diag([-C for C in c_values.values()]), matrix_ops.zeros((len(c_values), len(l_values))) )), # type: ignore
            matrix_ops.hstack(( matrix_ops.zeros((len(l_values), len(c_values))), matrix_ops.diag([L for L in l_values.values()]) ))
        ))

    Delta = element_incidence_matrix(c_values)
    A_tilde = nodal_analysis_coefficient_matrix(network, matrix_ops=matrix_ops, node_mapper=node_mapper, source_mapper=voltage_source_mapper)
    QS, QL = source_and_inductance_incidence_matrix(l_values)
    DQ = np.hstack((Delta.T, QL))
    Lambda = value_matrix(c_values, l_values)
    invLambda = matrix_ops.diag([1/L for L in matrix_ops.diag_vec(Lambda)]) # type: ignore

    inv_A_tilde = matrix_ops.inv(A_tilde)
    transformed_inv_A_tilde = DQ.T @ inv_A_tilde
    sorted_A_tilde = matrix_ops.inv(transformed_inv_A_tilde @ DQ)

    A = invLambda @ sorted_A_tilde
    C = transformed_inv_A_tilde.T @ sorted_A_tilde
    B = (-invLambda @ C.T) @ QS
    D = (inv_A_tilde - transformed_inv_A_tilde.T @ C.T) @ QS

    return A, B, C, D



import schemdraw, schemdraw.elements, schemdraw.util
from . import schemdraw_element_extension as extension

from . import Display as dsp

import numpy as np

from typing import Any
from collections import ChainMap
from enum import Enum
from abc import ABC

def round_node(node: schemdraw.util.Point) -> schemdraw.util.Point:
    def local_round(x):
        return round(x, ndigits=2)
    return schemdraw.util.Point((local_round(node.x), local_round(node.y)))

def get_nodes(element: schemdraw.elements.Element, n_labels: tuple[str, ...]=('start', 'end')) -> list[schemdraw.util.Point]:
    try:
        return [round_node(element.absanchors[n_label]) for n_label in n_labels]
    except KeyError:
        return []

def get_node_direction(node1: schemdraw.util.Point, node2: schemdraw.util.Point) -> tuple[int, int]:
    delta = node2 - node1
    delta_x = +1 if delta.x >= 0 else -1
    delta_y = +1 if delta.y >= 0 else -1
    return delta_x, delta_y

def is_reverse(element: schemdraw.elements.Element) -> bool:
    n1, n2 = get_nodes(element)
    dx, dy = get_node_direction(n1, n2)
    if dx < 0 or dy < 0:
        return True
    return False

def segments_of(element: schemdraw.elements.Element) -> list[schemdraw.segments.SegmentType]:
    return element.segments

class SwitchState(Enum):
    OPEN = False
    CLOSED = True

class Schematic(schemdraw.Drawing):
    def __init__(self, unit=7, **kwargs):
        super().__init__(unit=unit, **kwargs)

    def __getitem__(self, id: str) -> schemdraw.elements.Element:
        names = [e.name for e in self.elements if hasattr(e, 'name')]
        try:
            index = names.index(id)
        except ValueError:
            raise KeyError
        return self.elements[index]

    def clear_labels(self) -> None:
        self.elements = [e for e in self.elements if not isinstance(e, VoltageLabel) and not isinstance(e, CurrentLabel) and not isinstance(e, PowerLabel)]

    def draw(self, *args, **kwargs):
        if self.fig is not None:
            self.fig.clear()
        return super().draw(*args, **kwargs)

    def save_copy(self, fname: str, **kwargs) -> None:
        import copy
        cpy = copy.deepcopy(self)
        cpy.save(fname, **kwargs)

class SimpleCircuitElement(ABC):
    def __init__(self, *, name: str, reverse: bool = False):
        self._name = name
        self._reverse = reverse

    @property
    def name(self) -> str:
        return self._name

    @property
    def is_reverse(self) -> bool:
        return self._reverse

    @property
    def type(self) -> str:
        ...

def simple_circuit_element(element):
    class decorated_element(element, SimpleCircuitElement):
        def __init__(self, *args, **kwargs):
            element.__init__(self, *args, **kwargs)
            SimpleCircuitElement.__init__(self, name=kwargs.get('name', ''), reverse=kwargs.get('reverse', False))
    return decorated_element

@simple_circuit_element
class Element(schemdraw.elements.Element):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

@simple_circuit_element
@extension.source
class VoltageSource(schemdraw.elements.SourceV):
    def __init__(self, *args, name: str, V: float = float('nan'), show_name: bool = True, show_value: bool = True, reverse: bool = False, precision: int = 3, **kwargs):
        super().__init__(*args, reverse=not reverse, **kwargs)
        try:
            V = float(V)
        except ValueError:
            V = float('nan')
        if np.isnan(V):
            show_value = False
        self._V = V if not reverse else -V
        label = ''
        label += f'{name}' if show_name else ''
        label += '=' if  show_name and show_value else ''
        label += dsp.print_real(V, unit='V', precision=precision) if show_value else ''
        self.label(f'{label}', rotate=True, color=dsp.blue, loc='value_label', halign='center', valign='center')
        self.segments.append(extension.voltage_arrow())

    @property
    def V(self) -> float:
        return self._V

    @property
    def type(self) -> str:
        return 'voltage_source'

@simple_circuit_element
@extension.source
class ComplexVoltageSource(schemdraw.elements.SourceV):
    def __init__(self, *args, name: str, V: complex = complex('nan'), show_name: bool = True, show_value: bool = True, reverse: bool = False, precision: int = 3, **kwargs):
        super().__init__(*args, reverse=not reverse, **kwargs)
        try:
            V = complex(V)
        except ValueError:
            V = complex('nan')
        if np.isnan(V):
            show_value = False
        self._V = V if not reverse else -V
        label = ''
        label += f'{name}' if show_name else ''
        label += '=' if  show_name and show_value else ''
        label += dsp.print_complex(V, unit='V', precision=precision) if show_value else ''
        self.label(f'{label}', rotate=True, color=dsp.blue, loc='value_label', halign='center', valign='center')
        self.segments.append(extension.voltage_arrow())

    @property
    def V(self) -> complex:
        return self._V

    @property
    def type(self) -> str:
        return 'complex_voltage_source'

@simple_circuit_element
@extension.source
class CurrentSource(schemdraw.elements.SourceI):
    def __init__(self, *args, name: str, I: float = float('nan'), show_name: bool = True, show_value: bool = True, reverse: bool = False, precision: int = 3, **kwargs):
        super().__init__(*args, reverse=reverse, **kwargs)
        try:
            I = float(I)
        except ValueError:
            I = float('nan')
        if np.isnan(I):
            show_value = False
        self._I = I if not reverse else -I
        label = ''
        label += f'{name}' if show_name else ''
        label += '=' if  show_name and show_value else ''
        label += dsp.print_real(I, unit='A', precision=precision) if show_value else ''
        self.label(f'{label}', loc='i_label', ofst=(0, 0.4), rotate=True, color=dsp.red)
        self.segments.append(extension.current_arrow())

    @property
    def I(self) -> float:
        return self._I

    @property
    def type(self) -> str:
        return 'current_source'

@simple_circuit_element
@extension.source
class ComplexCurrentSource(schemdraw.elements.SourceI):
    def __init__(self, *args, name: str, I: complex = complex('nan'), show_name: bool = True, show_value: bool = True, reverse: bool = False, precision: int = 3, **kwargs):
        super().__init__(*args, reverse=reverse, **kwargs)
        try:
            I = complex(I)
        except ValueError:
            I = complex('nan')
        if np.isnan(I):
            show_value = False
        self._I = I if not reverse else -I
        label = ''
        label += f'{name}' if show_name else ''
        label += '=' if  show_name and show_value else ''
        label += dsp.print_complex(I, unit='A', precision=precision) if show_value else ''
        self.label(f'{label}', loc='i_label', ofst=(0, 0.4), rotate=True, color=dsp.red)
        self.segments.append(extension.current_arrow())

    @property
    def I(self) -> complex:
        return self._I

    @property
    def type(self) -> str:
        return 'complex_current_source'

@simple_circuit_element
@extension.resistor
class Resistor(schemdraw.elements.Resistor):
    def __init__(self, *args, name: str, R: float = float('nan'), show_name: bool = True, show_value: bool = True, reverse: bool = False, precision: int = 3, **kwargs):
        super().__init__(*args, reverse=reverse, **kwargs)
        try:
            R = float(R)
        except ValueError:
            R = float('nan')
        if np.isnan(R):
            show_value = False
        self._R = R
        label = ''
        label += f'{name}' if show_name else ''
        label += '=' if  show_name and show_value else ''
        label += dsp.print_resistance(self.R, precision=precision) if show_value else ''
        self.label(label, rotate=True, loc='value_label', halign='center')

    @property
    def R(self) -> float:
        return self._R

    @property
    def G(self) -> float:
        return 1/self._R

    @property
    def is_reverse(self) -> bool:
        return False

    @property
    def type(self) -> str:
        return 'resistor'

@simple_circuit_element
@extension.resistor
class Conductance(schemdraw.elements.Resistor):
    def __init__(self, *args, name: str, G: float = float('nan'), show_name: bool = True, show_value: bool = True, reverse: bool = False, precision: int = 3, **kwargs):
        super().__init__(*args, reverse=reverse, **kwargs)
        try:
            G = float(G)
        except ValueError:
            G = float('nan')
        if np.isnan(G):
            show_value = False
        self._G = G
        label = ''
        label += f'{name}' if show_name else ''
        label += '=' if  show_name and show_value else ''
        label += dsp.print_conductance(self.G, precision=precision) if show_value else ''
        self.label(label, rotate=True, loc='value_label', halign='center')

    @property
    def R(self) -> float:
        return 1/self._G

    @property
    def G(self) -> float:
        return self._G

    @property
    def is_reverse(self) -> bool:
        return False

    @property
    def type(self) -> str:
        return 'conductance'

@simple_circuit_element
@extension.resistor
class Impedance(schemdraw.elements.Resistor):
    def __init__(self, *args, name: str, Z: complex = complex('nan'), show_name: bool = True, show_value: bool = True, reverse: bool = False, precision: int = 3, **kwargs):
        super().__init__(*args, reverse=reverse, **kwargs)
        try:
            Z = complex(Z)
        except ValueError:
            Z = complex('nan')
        if np.isnan(Z):
            show_value = False
        self._Z = Z
        label = ''
        label += f'{name}' if show_name else ''
        label += '=' if  show_name and show_value else ''
        label += dsp.print_impedance(self.Z, precision=precision) if show_value else ''
        self.label(label, rotate=True, loc='value_label', halign='center')

    @property
    def Z(self) -> complex:
        return self._Z

    @property
    def Y(self) -> complex:
        return 1/self._Z

    @property
    def is_reverse(self) -> bool:
        return False

    @property
    def type(self) -> str:
        return 'impedance'

@simple_circuit_element
@extension.resistor
class Admittance(schemdraw.elements.Resistor):
    def __init__(self, *args, name: str, Y: complex = complex('nan'), show_name: bool = True, show_value: bool = True, reverse: bool = False, precision: int = 3, **kwargs):
        super().__init__(*args, reverse=reverse, **kwargs)
        try:
            Y = complex(Y)
        except ValueError:
            Y = complex('nan')
        if np.isnan(Y):
            show_value = False
        self._Y = Y
        label = ''
        label += f'{name}' if show_name else ''
        label += '=' if  show_name and show_value else ''
        label += dsp.print_impedance(self.Z, precision=precision) if show_value else ''
        self.label(label, rotate=True, loc='value_label', halign='center')

    @property
    def Z(self) -> complex:
        return 1/self._Y

    @property
    def Y(self) -> complex:
        return 1/self._Y

    @property
    def is_reverse(self) -> bool:
        return False

    @property
    def type(self) -> str:
        return 'admittance'

@simple_circuit_element
@extension.source
class ACVoltageSource(schemdraw.elements.SourceSin):
    def __init__(self, V: float, w: float, phi: float, name: str, *args, show_name: bool = True, show_value: bool = True, sin=False, deg=False, reverse=False, precision=3, **kwargs):
        super().__init__(*args, reverse=not reverse, **kwargs)
        self._V = V if not reverse else -V
        self._w = w
        self._phi = phi
        self._deg = deg
        self._sin = sin
        if self._sin:
            self._phi -= np.pi/2
        self.segments.append(extension.voltage_arrow())
        label = ''
        label += f'{name}' if show_name else ''
        label += '=' if  show_name and show_value else ''
        label += dsp.print_sinosoidal(V*np.exp(1j*self._phi), unit='V', precision=precision, w=w, deg=deg) if show_value else ''
        self.label(label, rotate=True, color=dsp.blue, loc='value_label', halign='center')

    @property
    def w(self) -> float:
        return self._w

    @property
    def phi(self) -> float:
        return self._phi

    @property
    def sin(self) -> bool:
        return self._sin

    @property
    def deg(self) -> bool:
        return self._deg

    @property
    def V(self) -> float:
        return self._V

    @property
    def type(self) -> str:
        return 'ac_voltage_source'

@simple_circuit_element
@extension.source
class ACCurrentSource(schemdraw.elements.SourceSin):
    def __init__(self, *args, I: float, w: float, phi: float, name: str, show_name: bool = True, show_value: bool = True, sin=False, deg=False, reverse=False, precision=3, **kwargs):
        super().__init__(*args, reverse=reverse, **kwargs)
        self._I = I if not reverse else -I
        self._w = w
        self._phi = phi
        self._deg = deg
        self._sin = sin
        if self._sin:
            self._phi -= np.pi/2
        label = ''
        label += f'{name}' if show_name else ''
        label += '=' if  show_name and show_value else ''
        label += dsp.print_sinosoidal(I*np.exp((1j*self._phi)), unit='A', precision=precision, w=w, deg=deg) if show_value else ''
        self.label(label, loc='i_label', ofst=(0, 0.4), rotate=True, color=dsp.red)
        self.segments.append(extension.current_arrow())

    @property
    def w(self) -> float:
        return self._w

    @property
    def phi(self) -> float:
        return self._phi

    @property
    def sin(self) -> bool:
        return self._sin

    @property
    def deg(self) -> bool:
        return self._deg

    @property
    def I(self) -> float:
        return self._I

    @property
    def type(self) -> str:
        return 'ac_current_source'

@extension.source
@simple_circuit_element
class RectVoltageSource(schemdraw.elements.SourceSquare):
    def __init__(self, V: float, w: float, phi: float, name: str, *args, sin=False, deg=False, reverse=False, **kwargs):
        super().__init__(*args, reverse=not reverse, **kwargs)
        self._V = V if not reverse else -V
        self._w = w
        self._phi = phi
        self._deg = deg
        self._sin = sin
        self.segments.append(extension.voltage_arrow())
        self.label(f'{name}', loc='value_label', halign='center', rotate=True, color=dsp.blue)

    @property
    def w(self) -> float:
        return self._w

    @property
    def phi(self) -> float:
        return self._phi

    @property
    def sin(self) -> bool:
        return self._sin

    @property
    def deg(self) -> bool:
        return self._deg

    @property
    def V(self) -> float:
        return self._V

    @property
    def type(self) -> str:
        return 'rect_voltage_source'

@simple_circuit_element
@extension.source
class TriangleVoltageSource(schemdraw.elements.SourceTriangle):
    def __init__(self, V: float, w: float, phi: float, name: str, *args, sin=False, deg=False, reverse=False, **kwargs):
        super().__init__(*args, reverse=not reverse, **kwargs)
        self._V = V if not reverse else -V
        self._w = w
        self._phi = phi
        self._deg = deg
        self._sin = sin
        self.segments.append(extension.voltage_arrow())
        self.label(f'{name}', loc='value_label', halign='center', rotate=True, color=dsp.blue)

    @property
    def w(self) -> float:
        return self._w

    @property
    def phi(self) -> float:
        return self._phi

    @property
    def sin(self) -> bool:
        return self._sin

    @property
    def deg(self) -> bool:
        return self._deg

    @property
    def V(self) -> float:
        return self._V

    @property
    def type(self) -> str:
        return 'tri_voltage_source'

@simple_circuit_element
@extension.source
class SawtoothVoltageSource(schemdraw.elements.Source):
    def __init__(self, V: float, w: float, phi: float, name: str, *args, sin=False, deg=False, reverse=False, **kwargs):
        super().__init__(*args, reverse=not reverse, **kwargs)
        self._V = V if not reverse else -V
        self._w = w
        self._phi = phi
        self._deg = deg
        self._sin = sin
        self.segments.append(schemdraw.segments.Segment([(0.3, 0.2), (0.7, 0), (0.3, 0), (0.7, -0.2), (0.3, -0.2)]))
        self.segments.append(extension.voltage_arrow())
        self.label(f'{name}', loc='value_label', halign='center', rotate=True, color=dsp.blue)

    @property
    def w(self) -> float:
        return self._w

    @property
    def phi(self) -> float:
        return self._phi

    @property
    def sin(self) -> bool:
        return self._sin

    @property
    def deg(self) -> bool:
        return self._deg

    @property
    def V(self) -> float:
        return self._V

    @property
    def type(self) -> str:
        return 'saw_voltage_source'

@simple_circuit_element
@extension.source
class RectCurrentSource(schemdraw.elements.SourceSquare):
    def __init__(self, I: float, w: float, phi: float, name: str, *args, sin=False, deg=False, reverse=False, **kwargs):
        super().__init__(*args, reverse=reverse, **kwargs)
        self._I = I if not reverse else -I
        self._w = w
        self._phi = phi
        self._deg = deg
        self._sin = sin
        self.segments.append(extension.current_arrow())
        self.label(f'{name}', loc='i_label', halign='center', rotate=True, color=dsp.red)

    @property
    def w(self) -> float:
        return self._w

    @property
    def phi(self) -> float:
        return self._phi

    @property
    def sin(self) -> bool:
        return self._sin

    @property
    def deg(self) -> bool:
        return self._deg

    @property
    def I(self) -> float:
        return self._I

    @property
    def type(self) -> str:
        return 'rect_current_source'

@simple_circuit_element
@extension.source
class TriangleCurrentSource(schemdraw.elements.SourceTriangle):
    def __init__(self, I: float, w: float, phi: float, name: str, *args, sin=False, deg=False, reverse=False, **kwargs):
        super().__init__(*args, reverse=reverse, **kwargs)
        self._I = I if not reverse else -I
        self._w = w
        self._phi = phi
        self._deg = deg
        self._sin = sin
        self.segments.append(extension.current_arrow())
        self.label(f'{name}', loc='i_label', halign='center', rotate=True, color=dsp.red)

    @property
    def w(self) -> float:
        return self._w

    @property
    def phi(self) -> float:
        return self._phi

    @property
    def sin(self) -> bool:
        return self._sin

    @property
    def deg(self) -> bool:
        return self._deg

    @property
    def I(self) -> float:
        return self._I

    @property
    def type(self) -> str:
        return 'tri_current_source'

@simple_circuit_element
@extension.source
class SawtoothCurrentSource(schemdraw.elements.Source):
    def __init__(self, I: float, w: float, phi: float, name: str, *args, sin=False, deg=False, reverse=False, **kwargs):
        super().__init__(*args, reverse=reverse, **kwargs)
        self._I = I if not reverse else -I
        self._w = w
        self._phi = phi
        self._deg = deg
        self._sin = sin
        self.segments.append(schemdraw.segments.Segment([(0.3, 0.2), (0.7, 0), (0.3, 0), (0.7, -0.2), (0.3, -0.2)]))
        self.segments.append(extension.current_arrow())
        self.label(f'{name}', loc='i_label', halign='center', rotate=True, color=dsp.red)

    @property
    def w(self) -> float:
        return self._w

    @property
    def phi(self) -> float:
        return self._phi

    @property
    def sin(self) -> bool:
        return self._sin

    @property
    def deg(self) -> bool:
        return self._deg

    @property
    def I(self) -> float:
        return self._I

    @property
    def type(self) -> str:
        return 'saw_current_source'

@simple_circuit_element
@extension.capacitor
class Capacitor(schemdraw.elements.Capacitor):
    def __init__(self, *args, name: str, C: float = float('nan'), show_name: bool = True, show_value: bool = True, reverse: bool = False, **kwargs):
        super().__init__(*args, reverse=reverse, **kwargs)
        try:
            C = float(C)
        except ValueError:
            C = float('nan')
        if np.isnan(C):
            show_value = False
        self._C = C
        label = ''
        label += f'{name}' if show_name else ''
        label += '=' if  show_name and show_value else ''
        label += dsp.print_capacitance(C) if show_value else ''
        self.label(label, rotate=True, loc='value_label', halign='center')

    @property
    def C(self) -> float:
        return self._C

    @property
    def is_reverse(self) -> bool:
        return False

    @property
    def type(self) -> str:
        return 'capacitor'

@simple_circuit_element
@extension.inductor
class Inductance(schemdraw.elements.Inductor):
    def __init__(self, *args, name: str, L: float = float('nan'), show_name: bool = True, show_value: bool = True, label_offset: float = 0.2, reverse: bool = False, **kwargs):
        super().__init__(*args, reverse=reverse, **kwargs)
        try:
            L = float(L)
        except ValueError:
            L = float('nan')
        if np.isnan(L):
            show_value = False
        self._L = L
        label = ''
        label += f'{name}' if show_name else ''
        label += '=' if  show_name and show_value else ''
        label += dsp.print_inductance(L) if show_value else ''
        self.label(label, rotate=True, loc='value_label', halign='center')

    @property
    def L(self) -> float:
        return self._L

    @property
    def is_reverse(self) -> bool:
        return False

    @property
    def type(self) -> str:
        return 'inductance'

@extension.linear_current_source
@simple_circuit_element
class RealCurrentSource(schemdraw.elements.compound.ElementCompound):
    def __init__(self, name: str, I: float, R: float, *args, **kwargs):
        self._I = I
        self._R = R
        self._name = name
        super().__init__(*args, **kwargs)

    def setup(self):
        zoom_resistor = 0.7
        cs = CurrentSource(I=self._I, name=self._name, d='r')
        res = Resistor(R=self._R, name='R', l=zoom_resistor, d='r')
        transform = schemdraw.transform.Transform(theta = 0, globalshift=(((-3-zoom_resistor)/2,-1.5)), localshift=(0, 0), zoom=zoom_resistor)
        res.segments = [s.xform(transform) for s in res.segments]
        res.anchors['value_label'] = (-1.5, -2.5)
        self.add(cs)
        self.add(res)
        self.add(l_up:=schemdraw.elements.Line(d='r', l=1).at(res.end))
        self.add(l_down:=schemdraw.elements.Line(d='l', l=1).at(res.start))
        self.add(schemdraw.elements.Line(d='u', l=1.5).at(l_up.end))
        self.add(schemdraw.elements.Line(d='u', l=1.5).at(l_down.end))

    @property
    def I(self) -> complex:
        return self._I

    @property
    def R(self) -> float:
        return self._R

    @property
    def G(self) -> float:
        return 1/self._R

    @property
    def type(self) -> str:
        return 'real_current_source'

@extension.linear_voltage_source
@simple_circuit_element
class RealVoltageSource(schemdraw.elements.compound.ElementCompound):
    def __init__(self, name: str, V: float, R: float, *args, **kwargs):
        self._V = V
        self._R = R
        self._name = name
        super().__init__(*args, **kwargs)

    def setup(self):
        zoom_resistor = 1 
        vs = VoltageSource(V=self._V, name=self._name, d='l')
        res = Resistor(R=self._R, name='R', l=zoom_resistor, d='l')
        transform_resistor = schemdraw.transform.Transform(theta = 0, globalshift=(0, 0), zoom=zoom_resistor)
        res.segments = [s.xform(transform_resistor) for s in res.segments]
        res.anchors['value_label'] = (0.5, -0.8)
        self.add(L:=schemdraw.elements.Line(d='l', l=1))
        self.add(res.at(L.end))
        self.add(vs)

    @property
    def V(self) -> complex:
        return self._V

    @property
    def R(self) -> float:
        return self._R

    @property
    def G(self) -> float:
        return 1/self._R

    @property
    def type(self) -> str:
        return 'real_voltage_source'

@simple_circuit_element
class Line(schemdraw.elements.lines.Line):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def name(self) -> str:
        return ''

    @property
    def type(self) -> str:
        return 'line'

@simple_circuit_element
@extension.lamp
class Lamp(schemdraw.elements.Lamp2):
    def __init__(self, V_ref: float, P_ref: float, name: str, *args, show_name: bool = True, show_value: bool = True, reverse: bool = False, precision: int = 3, **kwargs):
        super().__init__(*args, reverse=reverse, **kwargs)
        self._V_ref = V_ref
        self._P_ref = P_ref
        label = ''
        label += f'{name}' if show_name else ''
        label += '=' if  show_name and show_value else ''
        label += dsp.print_real(P_ref, unit='W', precision=precision) if show_value else ''
        self.label(label, rotate=show_value, loc='value_label', halign='center', valign='center')

    @property
    def V_ref(self) -> float:
        return self._V_ref

    @property
    def P_ref(self) -> float:
        return self._P_ref

    @property
    def R(self) -> float:
        return self._V_ref**2/self._P_ref

    @property
    def is_reverse(self) -> bool:
        return False

class LabeledLine(Line):
    def __init__(self, *args, name: str, **kwargs):
        super().__init__(*args, name=name, **kwargs)
        self._name = name

    @property
    def name(self) -> str:
        return self._name

    @property
    def type(self) -> str:
        return 'labeled_line'

    @property
    def is_reverse(self) -> bool:
        return False

@simple_circuit_element
class Node(schemdraw.elements.Element):
    def __init__(self, *args, name: str = '', **kwargs):
        super().__init__(*args, name=name, **kwargs)
        self.node_id = name
        self.params['theta'] = 0
        self.params['drop'] = (0, 0)
        self.anchors['start'] = (0, 0)
        self.anchors['center'] = (0, 0)
        self.anchors['end'] = (0, 0)
        self.anchors['NE'] = (0.5, 0.1)
        self.anchors['NW'] = (-0.5, 0.1)
        self.anchors['SE'] = (0.5, -0.3)
        self.anchors['SW'] = (-0.5, -0.3)

    @property
    def type(self) -> str:
        return 'node'

class LabelNode(Node):
    def __init__(self, id_loc : str | dict[str, str] = '', *args, name: str = '', show=True, **kwargs):
        super().__init__(*args, name=name, **kwargs)
        locations = {
            'W': {'loc': 'left', 'halign': 'right', 'valign': 'center'},
            'N': {'loc': 'top', 'halign': 'center', 'valign': 'bottom'},
            'NE': {'loc': 'NE', 'halign': 'left', 'valign': 'bottom'},
            'NW': {'loc': 'NW', 'halign': 'right', 'valign': 'bottom'},
            'E': {'loc': 'right', 'halign': 'left', 'valign': 'center'},
            'S': {'loc': 'bottom', 'halign': 'center', 'valign': 'top'},
            'SW': {'loc': 'SW', 'halign': 'right', 'valign': 'top'},
            'SE': {'loc': 'SE', 'halign': 'left', 'valign': 'top'}
        }
        self.segments.append(schemdraw.SegmentCircle((0, 0), 0.12, fill='black'))
        self.id_loc = {}
        if isinstance(id_loc, str):
            self.id_loc.update(locations.get(id_loc, {}))
        else:
            self.id_loc.update(id_loc)
        if show:
            self.show()

    def show(self):
        self.segments.append(schemdraw.SegmentCircle((0, 0), 0.12, fill='black'))
        self.bbox = self.get_bbox(includetext=False)
        self.label(f'{self.name}', **self.id_loc)

    @property
    def is_reverse(self) -> bool:
        return False

    @property
    def type(self) -> str:
        return 'label_node'

@simple_circuit_element
class Switch(schemdraw.elements.elements.Element2Term):
    def __init__(self, name: str, *args, state: SwitchState = SwitchState.OPEN, **kwargs):
        super().__init__(*args, **kwargs)
        self.state = state
        self._name = name
        self.segments.append(schemdraw.Segment([(0, 0), schemdraw.elements.elements.gap, (1, 0)]))
        self.segments.append(schemdraw.SegmentCircle((schemdraw.elements.switches.sw_dot_r, 0), schemdraw.elements.switches.sw_dot_r, fill='bg', zorder=3))
        self.segments.append(schemdraw.SegmentCircle((1-schemdraw.elements.switches.sw_dot_r, 0), schemdraw.elements.switches.sw_dot_r, fill='bg', zorder=3))
        self.segments.append(schemdraw.Segment([(0,0)]))
        if self.state == SwitchState.OPEN:
            self.open()
        if self.state == SwitchState.CLOSED:
            self.close()

    @property
    def name(self) -> str:
        return self._name

    @property
    def is_reverse(self) -> bool:
        return False

    @property
    def type(self) -> str:
        return 'switch'

    def open(self) -> None:
        self.state = SwitchState.OPEN
        self.segments.pop()
        self.segments.append(schemdraw.Segment([(schemdraw.elements.switches.sw_dot_r, 0), (1, 0.6)]))

    def close(self) -> None:
        self.state = SwitchState.CLOSED
        self.segments.pop()
        self.segments.append(schemdraw.Segment([(schemdraw.elements.switches.sw_dot_r, 0), (1, schemdraw.elements.switches.sw_dot_r*1.2)]))

    def toggle(self) -> None:
        if self.state == self.state.OPEN:
            self.close()
        if self.state == self.state.CLOSED:
            self.open()

class Ground(Node):
    def __init__(self, *args, name: str = '0', **kwargs):
        super().__init__(*args, name=name, **kwargs)
        gndgap = 0.12
        gnd_lead = 0.4
        resheight = schemdraw.elements.twoterm.resheight
        gap = schemdraw.elements.elements.gap
        self.segments.append(schemdraw.Segment(
            [(0, 0), (0, -gnd_lead), (-resheight, -gnd_lead),
             (resheight, -gnd_lead), gap, (-resheight*.7, -gndgap-gnd_lead),
             (resheight*.7, -gndgap-gnd_lead), gap,
             (-resheight*.2, -gndgap*2-gnd_lead),
             (resheight*.2, -gndgap*2-gnd_lead)]))

    @property
    def type(self) -> str:
        return 'ground'

    def up(self, _: float) -> schemdraw.elements.Element:
        return super().left()

    def down(self, _: float) -> schemdraw.elements.Element:
        return super().right()

    def left(self, _: float) -> schemdraw.elements.Element:
        return super().down()

    def right(self, _: float) -> schemdraw.elements.Element:
        return super().up()

class VoltageLabel(schemdraw.elements.CurrentLabel):
    def __init__(self, at: schemdraw.elements.Element, vlabel: str = '', label_loc: str = 'bottom', reverse: bool = False, color: str | tuple[float, float, float] = dsp.blue, **kwargs):
        self.params = ChainMap({'headwidth': 0.3, 'headlength': 0.4, 'color': color}, self.params)
        if isinstance(at, RealVoltageSource):
            kwargs.update({'length': kwargs.get('length', 4)})
        super().__init__(reverse=reverse, **kwargs)
        try:
            self.at(at.v_label)
            self.theta(at.transform.theta)
        except AttributeError:
            self.at(at.center)
        rotate = kwargs.get('rotate', True)
        if rotate == True and at.transform.theta == 270:
            rotate = 90
        self.label(vlabel, rotate=rotate, loc=label_loc, ofst=(0, -0.1))

class CurrentLabel(schemdraw.elements.CurrentLabelInline):
    def __init__(self, at: schemdraw.elements.Element, ilabel: str = '', **kwargs):
        kwargs.update(i_label_args.get(type(at), {}))
        kwargs.update({'color': kwargs.get('color', dsp.red)})
        kwargs.update({'headlength': kwargs.get('headlength', 0.4)})
        kwargs.update({'headwidth': kwargs.get('headwidth', 0.3)})
        totlen = at.params.get('l', at.params.get('unit', 7))
        kwargs.update({'ofst': totlen/4-kwargs['headlength']/2})
        start = kwargs.get('start', True)
        reverse = kwargs.get('reverse', False)
        if isinstance(at, RealVoltageSource) or isinstance(at, RealCurrentSource): # when replacing CurrentLabelInline this dependency may be removed
            reverse = not reverse
        kwargs.update({'start' : start, 'reverse' : reverse})
        super().__init__(**kwargs)
        self.at(at)
        self.label(ilabel)

class PowerLabel(schemdraw.elements.Label):
    def __init__(self, at: schemdraw.elements.Element, plabel: str = '', offset: float = 0, **kwargs):
        kwargs.update({'color': kwargs.get('color', dsp.green)})
        super().__init__(**kwargs)
        try:
            self.at(at.s_label)
            self.theta(at.transform.theta)
        except AttributeError:
            self.at(at.center)
        rotate = kwargs.get('rotate', True)
        if rotate == True and at.transform.theta == 270:
            rotate = 90
        self.label(plabel, rotate=rotate, halign='center', ofst=(0, offset))

i_label_args : dict[Any, dict[str, float]] = {
    Resistor : {'ofst' : 0},
    Conductance : {'ofst' : 1.4},
    Impedance : {'ofst' : 1.4},
    Admittance : {'ofst' : 1.4},
    Capacitor : {'ofst' : 1.4},
    Inductance : {'ofst' : 1.4},
    VoltageSource : {'ofst' : -2.8},
    ComplexVoltageSource : {'ofst' : -2.8},
    CurrentSource : {'ofst' : -2.8},
    ComplexCurrentSource : {'ofst' : -2.8},
    ACVoltageSource : {'ofst' : -3.8},
    RealVoltageSource: {'ofst' : -0.8},
    RealCurrentSource: {'ofst' : 1.4}
}


from CircuitCalculator.Network.NodalAnalysis.bias_point_analysis import nodal_analysis_bias_point_solver
from CircuitCalculator.Network.loaders import load_network_from_json
from CircuitCalculator.Network.NodalAnalysis.node_analysis import open_circuit_impedance
import numpy as np
from pathlib import Path

json_root = Path('.') / 'examples' / 'test-networks' / '01_json-network'

def test_network_1_with_advanced_nodal_analysis() -> None:
    network = load_network_from_json(str(json_root / 'example_network_1.json'))
    solution = nodal_analysis_bias_point_solver(network)
    np.testing.assert_almost_equal(solution.get_voltage('Uq'), 1.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('Uq'), -1.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R'), 1.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('R'), 1.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('0'), 0.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('1'), 1.00, decimal=2)

def test_network_2_with_advanced_nodal_analysis() -> None:
    network = load_network_from_json(str(json_root / 'example_network_2.json'))
    solution = nodal_analysis_bias_point_solver(network)
    np.testing.assert_almost_equal(solution.get_voltage('Iq'), -1.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('Iq'), 1.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R'), 1.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('R'), 1.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('0'), 0.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('1'), 1.00, decimal=2)

def test_network_3_with_advanced_nodal_analysis() -> None:
    network = load_network_from_json(str(json_root / 'example_network_3.json'))
    solution = nodal_analysis_bias_point_solver(network)
    np.testing.assert_almost_equal(solution.get_voltage('R1'), 7.69, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R2'), 15.38, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('I1'), -23.08, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('R1'), 0.77, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('R2'), 0.77, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('I1'), -0.77, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('0'), 0.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('1'), 23.08, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('2'), 15.38, decimal=2)

def test_network_4_with_advanced_nodal_analysis() -> None:
    network = load_network_from_json(str(json_root / 'example_network_4.json'))
    solution = nodal_analysis_bias_point_solver(network)
    np.testing.assert_almost_equal(solution.get_voltage('R1'), 1.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R2'), 0.40, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R3'), 0.60, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('U1'), 1.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('R1'), 0.10, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('R2'), 0.02, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('R3'), 0.02, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('U1'), -0.12, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('0'), 0.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('1'), 1.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('2'), 0.60, decimal=2)

def test_network_5_with_advanced_nodal_analysis() -> None:
    network = load_network_from_json(str(json_root / 'example_network_5.json'))
    solution = nodal_analysis_bias_point_solver(network)
    np.testing.assert_almost_equal(solution.get_voltage('R1'), 0.56, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R2'), 0.44, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R3'), 1.04, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R4'), 1.39, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R5'), -2.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('U1'), 1.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('U2'), 2.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('R1'), 0.056, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('R2'), 0.022, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('R3'), 0.035, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('R4'), 0.035, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('R5'), -0.04, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('U1'), -0.056, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('U2'), -0.075, decimal=3)
    np.testing.assert_almost_equal(solution.get_potential('0'), 0.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('1'), 1.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('2'), 0.44, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('3'), -0.61, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('4'), -2.00, decimal=2)
    
def test_network_6_with_advanced_nodal_analysis() -> None:
    network = load_network_from_json(str(json_root / 'example_network_6.json'))
    solution = nodal_analysis_bias_point_solver(network)
    np.testing.assert_almost_equal(solution.get_voltage('R1'), -0.52, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R2'), -0.48, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R3'), 2.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R4'), -1.10, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R5'), -1.38, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('U1'), 1.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('U2'), 2.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('R1'), -0.052, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('R2'), -0.024, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('R3'), 0.067, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('R4'), -0.028, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('R5'), -0.028, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('U1'), -0.052, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('U2'), -0.094, decimal=3)
    np.testing.assert_almost_equal(solution.get_potential('0'), 0.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('1'), -1.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('2'), -0.48, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('3'), -2.48, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('4'), -1.38, decimal=2)

def test_network_7_with_advanced_nodal_analysis() -> None:
    network = load_network_from_json(str(json_root / 'example_network_7.json'))
    solution = nodal_analysis_bias_point_solver(network)
    np.testing.assert_almost_equal(solution.get_voltage('R1'), -1.56, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R2'), 2.14, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R3'), 1.44, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R4'), 0.30, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('U1'), 1.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('U2'), 2.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('U3'), 3.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('I4'), -2.7, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('R1'), -0.156, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('R2'), 0.107, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('R3'), 0.048, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('R4'), 0.0075, decimal=4)
    np.testing.assert_almost_equal(solution.get_current('U1'), -0.156, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('U2'), -0.048, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('U3'), -0.0075, decimal=4)
    np.testing.assert_almost_equal(solution.get_current('I4'), 0.1, decimal=3)
    np.testing.assert_almost_equal(solution.get_potential('0'), 0.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('1'), -1.56, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('2'), 0.44, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('3'), -3.71, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('4'), -0.71, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('5'), -1.00, decimal=2)

def test_network_8_with_advanced_nodal_analysis() -> None:
    network = load_network_from_json(str(json_root / 'example_network_8.json'))
    Rges = open_circuit_impedance(network, '1', '2').real
    np.testing.assert_almost_equal(Rges, 4.66, decimal=2)

def test_network_9_with_advanced_nodal_analysis() -> None:
    network = load_network_from_json(str(json_root / 'example_network_9.json'))
    solution = nodal_analysis_bias_point_solver(network)
    np.testing.assert_almost_equal(solution.get_voltage('Vs1'), 1.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('Vs2'), 2.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R1'), 3.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('Vs1'), -0.30, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('Vs2'), -0.30, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('R1'), 0.30, decimal=3)
    np.testing.assert_almost_equal(solution.get_potential('0'), 0.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('1'), 1.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('2'), 3.00, decimal=2)

def test_network_10_with_advanced_nodal_analysis() -> None:
    network = load_network_from_json(str(json_root / 'example_network_10.json'))
    solution = nodal_analysis_bias_point_solver(network)
    np.testing.assert_almost_equal(solution.get_voltage('R1'), 1.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R2'), 2.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R3'), -0.75, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R4'), -4.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R5'), 3.75, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('U1'), 1.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('U2'), 2.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('I3'), -7.75, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('R1'), 0.10, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('R2'), 0.10, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('R3'), -0.025, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('R4'), -0.10, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('R5'), 0.075, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('U1'), -0.075, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('U2'), -0.075, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('I3'), 0.10, decimal=3)
    np.testing.assert_almost_equal(solution.get_potential('0'), 0.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('1'), 1.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('2'), 3.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('3'), -0.75, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('4'), 7.00, decimal=2)

def test_network_11_with_advanced_nodal_analysis() -> None:
    network = load_network_from_json(str(json_root / 'example_network_11.json'))
    solution = nodal_analysis_bias_point_solver(network)
    np.testing.assert_almost_equal(solution.get_voltage('R1'), 0, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R2'), 20, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R3'), -20, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('R4'), 100, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('Iq'), 0, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('Uq'), 120, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('R1'), 0, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('R2'), 4, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('R3'), -1, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('R4'), 5, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('Iq'), 4, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('Uq'), -5, decimal=3)
    np.testing.assert_almost_equal(solution.get_potential('0'), 0.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('1'), 0.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('2'), -20.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('3'), 100.00, decimal=2)

def test_network_12_with_advanced_nodal_analysis() -> None:
    network = load_network_from_json(str(json_root / 'example_network_12.json'))
    solution = nodal_analysis_bias_point_solver(network)
    np.testing.assert_almost_equal(solution.get_voltage('Z1'), 1-2j, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('Z2'), 0.4-0.8j, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('Z3'), 0.6-1.2j, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('Us'), 1-2j, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('Z1'), -0.05-0.15j, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('Z2'), 0.02-0.04j, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('Z3'), 0.02-0.04j, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('Us'), 0.03+0.19j, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('0'), 0.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('1'), 1-2j, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('2'), 0.6-1.2j, decimal=2)

def test_network_13_with_advanced_nodal_analysis() -> None:
    network = load_network_from_json(str(json_root / 'example_network_13.json'))
    solution = nodal_analysis_bias_point_solver(network)
    np.testing.assert_almost_equal(solution.get_voltage('R1'), 0.182, decimal=3)
    np.testing.assert_almost_equal(solution.get_voltage('R2'), 0.182, decimal=3)
    np.testing.assert_almost_equal(solution.get_voltage('R3'), 0.818, decimal=3)
    np.testing.assert_almost_equal(solution.get_voltage('Vs'), 1, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('R1'), 0.0182, decimal=4)
    np.testing.assert_almost_equal(solution.get_current('R2'), 0.0091, decimal=4)
    np.testing.assert_almost_equal(solution.get_current('R3'), 0.0273, decimal=4)
    np.testing.assert_almost_equal(solution.get_current('Vs'), -0.0273, decimal=3)
    np.testing.assert_almost_equal(solution.get_potential('0'), 0.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('1'), 1.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('2'), 0.818, decimal=3)

def test_network_14_with_advanced_nodal_analysis() -> None:
    network = load_network_from_json(str(json_root / 'example_network_14.json'))
    solution = nodal_analysis_bias_point_solver(network)
    np.testing.assert_almost_equal(solution.get_voltage('R1'), 0.4, decimal=3)
    np.testing.assert_almost_equal(solution.get_voltage('R2'), 0.4, decimal=3)
    np.testing.assert_almost_equal(solution.get_voltage('Vs'), -0.4, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('R1'), 0.04, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('R2'), 0.02, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('Vs'), -0.06, decimal=3)
    np.testing.assert_almost_equal(solution.get_potential('0'), 0.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('1'), 0.40, decimal=2)

def test_network_15_with_advanced_nodal_analysis() -> None:
    network = load_network_from_json(str(json_root / 'example_network_15.json'))
    solution = nodal_analysis_bias_point_solver(network)
    np.testing.assert_almost_equal(solution.get_voltage('R1'), 0.36, decimal=3)
    np.testing.assert_almost_equal(solution.get_voltage('Vs'), 1.0, decimal=2)
    np.testing.assert_almost_equal(solution.get_voltage('sc'), 0.0, decimal=2)
    np.testing.assert_almost_equal(solution.get_current('R1'), 0.036, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('Vs'), -0.048, decimal=3)
    np.testing.assert_almost_equal(solution.get_current('sc'), 0.004, decimal=3)
    np.testing.assert_almost_equal(solution.get_potential('0'), 0.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('1'), 1.00, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('2'), 0.64, decimal=2)
    np.testing.assert_almost_equal(solution.get_potential('3'), 0.64, decimal=2)


from .norten_thevenin_elements import NortenTheveninElement, SymbolicNortenElement, SymbolicTheveninElement
import sympy as sp

def impedance(name : str, Z : sp.Symbol) -> NortenTheveninElement:
    return SymbolicNortenElement(Z=Z, V=sp.sympify(0), name=name, type='impedance')

def admittance(name : str, Y : sp.Symbol) -> NortenTheveninElement:
    return SymbolicTheveninElement(Y=Y, I=sp.sympify(0), name=name, type='admittance')

def resistor(name : str, R : sp.Symbol) -> NortenTheveninElement:
    return SymbolicNortenElement(Z=R, V=sp.sympify(0), name=name, type='resistor')

def conductor(name : str, G : sp.Symbol) -> NortenTheveninElement:
    return SymbolicTheveninElement(Y=G, I=sp.sympify(0), name=name, type='conductor')

def voltage_source(name : str, V : sp.Symbol, Z : sp.Symbol = sp.sympify(0)) -> NortenTheveninElement:
    return SymbolicNortenElement(V=V, Z=sp.sympify(Z), name=name, type='voltage_source')

def current_source(name : str, I : sp.Symbol, Y : sp.Symbol = sp.sympify(0)) -> NortenTheveninElement:
    return SymbolicTheveninElement(I=I, Y=sp.sympify(Y), name=name, type='current_source')

def open_circuit(name : str) -> NortenTheveninElement:
    return SymbolicTheveninElement(I=sp.sympify(0), Y=sp.sympify(0), name=name, type='open_circuit')

def short_circuit(name : str) -> NortenTheveninElement:
    return SymbolicNortenElement(V=sp.sympify(0), Z=sp.sympify(0), name=name, type='short_circuit')



import numpy as np
import sympy as sp
from .circuit import Circuit, transform_circuit, transform_symbolic_circuit
from ..Network.NodalAnalysis import node_analysis as na
from ..Network import matrix_operations as mo

def open_circuit_impedance(circuit: Circuit, node1: str, node2: str, w: np.ndarray = np.array([0])) -> np.ndarray:
    return np.array([na.open_circuit_impedance(transform_circuit(circuit, w0), node1, node2) for w0 in w])

def element_impedance(circuit: Circuit, element_id: str, w: np.ndarray = np.array([0])) -> np.ndarray:
    return np.array([na.element_impedance(transform_circuit(circuit, w0), element_id) for w0 in w])

def open_circuit_dc_resistance(circuit: Circuit, node1: str, node2: str) -> float:
    return open_circuit_impedance(circuit, node1, node2, w=np.array([0]))[0].real

def element_dc_resistance(circuit: Circuit, element_id: str) -> float:
    return element_impedance(circuit, element_id, w=np.array([0]))[0].real

def symbolic_open_circuit_impedance(circuit: Circuit, node1: str, node2: str, s: sp.Symbol = sp.Symbol('s', complex=True)) -> mo.symbolic:
    return na.open_circuit_impedance(transform_symbolic_circuit(circuit, s=s), node1, node2, matrix_ops = mo.SymPyMatrixOperations())

def symbolic_element_impedance(circuit: Circuit, element_id: str, s: sp.Symbol = sp.Symbol('s', complex=True)) -> mo.symbolic:
    return na.element_impedance(transform_symbolic_circuit(circuit, s=s), element_id, matrix_ops = mo.SymPyMatrixOperations())

def symbolic_open_circuit_dc_resistance(circuit: Circuit, node1: str, node2: str) -> mo.symbolic:
    return symbolic_open_circuit_impedance(circuit, node1, node2, s=sp.sympify(0))

def symbolic_element_dc_resistance(circuit: Circuit, element_id: str) -> mo.symbolic:
    return symbolic_element_impedance(circuit, element_id, s=sp.sympify(0))


from . import layout
from ..Circuit.solution import TimeDomainSolution, TimeDomainFunction
import numpy as np
import functools
from typing import Callable, TypedDict, Dict

def plot_timeseries_fcn(ax, x_fcn, tmin, tmax, N_samples, **kwargs) -> None:
    t = np.linspace(tmin, tmax, N_samples)
    ax.plot(t, x_fcn(t), **kwargs)

def plot_timeseries_by_id(
          id:str,
          t_min:float=0,
          t_max:float=0,
          N_samples:int=500,
          **kwargs) -> layout.PlotFcn:
    def plot_timeseries(
              fig:layout.Figure,
              ax:layout.Axes,
              ts_fcn:Callable[[str], TimeDomainFunction]=lambda _: lambda t: np.zeros(np.size(t)),
              label_fcn:Callable[[str], str]=lambda _: '',
              **_) -> layout.FigureAxes:
        x_min, x_max = t_min, t_max
        if t_min >= t_max:
            x_min, x_max = ax[0].get_xlim()
        if 'label' not in kwargs.keys():
            kwargs.update({'label':label_fcn(id)})
        plot_timeseries_fcn(ax[0], ts_fcn(id), x_min, x_max, N_samples, **kwargs)
        return fig, ax
    return plot_timeseries

def plot_timeseries_by_fcn(
          ts_fcn:TimeDomainFunction,
          t_min:float=0,
          t_max:float=0,
          N_samples:int=500,
          **kwargs) -> layout.PlotFcn:
    def plot_timeseries(fig:layout.Figure, ax:layout.Axes, **_) -> layout.FigureAxes:
        x_min, x_max = t_min, t_max
        if t_min >= t_max:
            x_min, x_max = ax[0].get_xlim()
        plot_timeseries_fcn(ax[0], ts_fcn, x_min, x_max, N_samples, **kwargs)
        return fig, ax
    return plot_timeseries

class PlotTimeSeriesProperties(TypedDict):
     ts_fcn: Callable[[str], TimeDomainFunction]
     label_fcn: Callable[[str], str]
     ylabel: str

def plot_timeseries_factory(
        *args,
        type:str='default',
        solution:TimeDomainSolution,
        tmax:float=-1,
        tmin:float=0,
        xlabel:str='t→',
        ylabel:str='') -> layout.PlotFcn:

    time_series_configurations: Dict[str, PlotTimeSeriesProperties] = {
        'voltage' : {'ts_fcn': solution.get_voltage, 'label_fcn': lambda id: f'V({id})', 'ylabel': 'v(t)→'},
        'current' : {'ts_fcn': solution.get_current, 'label_fcn': lambda id: f'I({id})', 'ylabel': 'i(t)→'},
        'power' : {'ts_fcn': solution.get_power, 'label_fcn': lambda id: f'P({id})', 'ylabel': 'p(t)→'},
        'default' : {'ts_fcn': lambda _: lambda t: np.zeros(np.size(t)), 'label_fcn': lambda _: '', 'ylabel': ''}
    }

    ts_properties = time_series_configurations.get(type, time_series_configurations['default'])
    ylabel = ylabel if len(ylabel) > 0 else ts_properties['ylabel']

    @layout.legend()
    @layout.grid()
    @layout.xlim_bottom(xmin=tmin, xmax=tmax)
    @layout.xlabel_bottom(xlabel)
    def plot_timeseries(fig:layout.Figure, ax:layout.Axes) -> layout.FigureAxes:
        ax[-1].set_ylabel(ylabel)
        updated_args = [functools.partial(a, ts_fcn=ts_properties['ts_fcn'], label_fcn=ts_properties['label_fcn']) for a in args]
        layout.apply_plt_fcn(fig, ax, *updated_args)
        return fig, ax
    return plot_timeseries

def time_domain_plot(*args, layout_fcn:layout.Layout=layout.figure_default, **kwargs) -> layout.FigureAxes:
        plot_timeseries = plot_timeseries_factory(*args, type='default', **kwargs)
        return plot_timeseries(*layout_fcn())

def steady_state_voltage_timedomain_analysis(*args, layout_fcn:layout.Layout=layout.figure_default, **kwargs) -> layout.FigureAxes:
        plot_timeseries = plot_timeseries_factory(*args, type='voltage', **kwargs)
        return plot_timeseries(*layout_fcn())

def steady_state_current_timedomain_analysis(*args, layout_fcn:layout.Layout=layout.figure_default, **kwargs) -> layout.FigureAxes:
        plot_timeseries = plot_timeseries_factory(*args, type='current', **kwargs)
        return plot_timeseries(*layout_fcn())

def steady_state_instantaneous_power_analysis(*args, layout_fcn:layout.Layout=layout.figure_default, **kwargs) -> layout.FigureAxes:
        plot_timeseries = plot_timeseries_factory(*args, type='power', **kwargs)
        return plot_timeseries(*layout_fcn())


from . import layout
from ..Circuit.solution import FrequencyDomainSolution, FrequencyDomainSeries
import functools
from typing import Callable, Dict, TypedDict
import numpy as np

def plot_nyquist_by_id(id:str, **kwargs) -> layout.PlotFcn:
    def plot_pointer(fig:layout.Figure, ax:layout.Axes, *, fd_fcn:Callable[[str], tuple[list[float], list[complex]]]=lambda _: ([0], [0+0j]), label_fcn:Callable[[str], str]=lambda _: '') -> layout.FigureAxes:
        w, z = fd_fcn(id)
        if 'label' not in kwargs.keys():
            kwargs.update({'label':label_fcn(id)})
        if 'marker' not in kwargs.keys():
            kwargs.update({'marker':'o'})
        kwargs.update({'linestyle': ''})

        ax[0].plot(np.real(z), np.imag(z), **kwargs)
        return fig, ax
    return plot_pointer

def plot_nyquist_by_value(z:list[complex], **kwargs) -> layout.PlotFcn:
    def plot_pointer(fig:layout.Figure, ax:layout.Axes, **_) -> layout.FigureAxes:
        ax[0].plot(np.real(z), np.imag(z), **kwargs)
        return fig, ax
    return plot_pointer

class NyquistDiagramProperties(TypedDict):
    fd_fcn: Callable[[str], FrequencyDomainSeries]
    label_fcn: Callable[[str], str]
    xlabel:str
    ylabel:str

def plot_nyquist_diagram_factory(
        *args,
        type:str='default',
        solution:FrequencyDomainSolution,
        ax_lim:tuple[float, float, float, float]=(-1, 1, -1, 1),
        xlabel:str='',
        ylabel:str='') -> layout.PlotFcn:

    nyquist_diagram_configuration: Dict[str, NyquistDiagramProperties] = {
        'default' : {'fd_fcn' : lambda _: (np.zeros(1), np.zeros(1)), 'label_fcn' : lambda _: '', 'xlabel' : 'Re→', 'ylabel' : 'Im→'},
        'voltage' : {'fd_fcn' : solution.get_voltage, 'label_fcn' : lambda id: f'V({id})', 'xlabel' : 'Re{V}→', 'ylabel' : 'Im{V}→'},
        'current' : {'fd_fcn' : solution.get_current, 'label_fcn' : lambda id: f'I({id})', 'xlabel' : 'Re{I}→', 'ylabel' : 'Im{I}→'},
        'power' : {'fd_fcn' : solution.get_power, 'label_fcn' : lambda id: f'S({id})', 'xlabel' : 'P→', 'ylabel' : 'Q→'}
    }
    nd_properties = nyquist_diagram_configuration.get(type, nyquist_diagram_configuration['default'])
    xlabel = xlabel if len(xlabel) > 0 else nd_properties['xlabel']
    ylabel = ylabel if len(ylabel) > 0 else nd_properties['ylabel']

    @layout.legend()
    @layout.grid()
    @layout.nyquist_like_plot(ax_lim=ax_lim, xlabel=xlabel, ylabel=ylabel)
    def plot_nyquist_diagram(fig:layout.Figure, ax:layout.Axes) -> layout.FigureAxes:
        new_args = [functools.partial(a, fd_fcn=nd_properties['fd_fcn'], label_fcn=nd_properties['label_fcn']) for a in args]
        layout.apply_plt_fcn(fig, ax, *new_args)
        return fig, ax
    return plot_nyquist_diagram

def nyquist_plot(*args, layout_fcn:layout.Layout=layout.figure_default, **kwargs) -> layout.FigureAxes:
    plot_nyquist_diagram = plot_nyquist_diagram_factory(*args, type='default', **kwargs)
    return plot_nyquist_diagram(*layout_fcn())

def nyquist_voltage_analysis(*args, layout_fcn:layout.Layout=layout.figure_default, **kwargs) -> layout.FigureAxes:
    plot_nyquist_diagram = plot_nyquist_diagram_factory(*args, type='voltage', **kwargs)
    return plot_nyquist_diagram(*layout_fcn())

def nyquist_current_analysis(*args, layout_fcn:layout.Layout=layout.figure_default, **kwargs) -> layout.FigureAxes:
    plot_nyquist_diagram = plot_nyquist_diagram_factory(*args, type='current', **kwargs)
    return plot_nyquist_diagram(*layout_fcn())

def nyquist_power_analysis(*args, layout_fcn:layout.Layout=layout.figure_default, **kwargs) -> layout.FigureAxes:
    plot_nyquist_diagram = plot_nyquist_diagram_factory(*args, type='power', **kwargs)
    return plot_nyquist_diagram(*layout_fcn())
